"""
.. _serp:

Import Search Engine Results Pages (SERPs) for Google and YouTube
=================================================================
Analyzing a single SERP is like getting one person to fill out a questionnaire
and calling it a survey.

Just like surveys, SERPs need to be collected in large-enough numbers that are
representative of the industry/market you want to understand. This is the main
feature of the ``serp_`` functions. They allow you to get the SERPs for a list
of queries, across several dimensions (like country, search type, start
position, and so on).

There are many parameters that can be used, and you can supply a list for each.
The function will get the SERPs for the *product* of all those lists. For
example, let's say you you provide the following arguments to the
:func:`serp_goog` function:

* `q`: ['serp tools', 'best serp tools', 'serp tool reviews']
* `gl`: ['us', 'ca', 'uk', 'au', 'nz']
* `start`: [1, 11, 21]

The function will produce:
3 (queries) x 5 (countries) x 3 (start positions) = 45 requests

You typically get ten results each, so in this case you would get 450 rows of
data.

All this is done in with one line of code. The result is a single DataFrame
with a row for each result, and columns for each attribute (title, snippet,
etc.), as well as meta data columns, like `queryTime` and the parameters you
selected (`q`, `gl`, and `start` in this case).


Before being able to run queries using :func:`serp_goog`, you will need to set
up some credentials as follows (you don't need a custom search engine for
:func:`serp_youtube`):

* `Create a custom search engine <https://cse.google.com/>`_: At first, you might be asked to enter a site to search. Enter any domain, then go to the control panel and remove it. Make sure you enable "Search the entire web" and image search. You will also need to get your search engine ID, which you can find on the control panel page.

* `Enable the custom search API <https://console.cloud.google.com/apis/library/customsearch.googleapis.com?pli=1>`_: The service will allow you to retrieve and display search results from your custom search engine programmatically. You will need to create a project for this first.

* `Create credentials for this project <https://console.developers.google.com/apis/api/customsearch.googleapis.com/credentials>`_: so you can get your key.

* `Enable billing for your project <https://console.cloud.google.com/billing/projects>`_ if you want to run more than 100 queries per day. The first 100 queries are free; then for each additional 1,000 queries, you pay $5.


"""
__all__ = ['SERP_GOOG_VALID_VALS', 'YOUTUBE_TOPIC_IDS',
           'YOUTUBE_VID_CATEGORY_IDS', 'serp_goog', 'serp_youtube',
           'set_logging_level', 'youtube_channel_details',
           'youtube_video_details']

import datetime
import logging
from itertools import product

import pandas as pd

if int(pd.__version__[0]) >= 1:
    from pandas import json_normalize
else:
    from pandas.io.json import json_normalize

import requests

SERP_GOOG_LOG_FMT = ('%(asctime)s | %(levelname)s | %(filename)s:%(lineno)d '
                     '| %(funcName)s | %(message)s')
logging.basicConfig(format=SERP_GOOG_LOG_FMT)


##############################################################################
# Google variables
##############################################################################


SERP_GOOG_VALID_VALS = dict(
    fileType={
        'bas', 'c', 'cc', 'cpp', 'cs', 'cxx', 'doc', 'docx', 'dwf', 'gpx',
        'h', 'hpp', 'htm', 'html', 'hwp', 'java', 'kml', 'kmz', 'odp', 'ods',
        'odt', 'pdf', 'pl', 'ppt', 'pptx', 'ps', 'py', 'rtf', 'svg', 'swf',
        'tex', 'text', 'txt', 'wap', 'wml', 'xls', 'xlsx', 'xml',
    },

    c2coff={0, 1},

    cr={
        'countryAF', 'countryAL', 'countryDZ', 'countryAS', 'countryAD',
        'countryAO', 'countryAI', 'countryAQ', 'countryAG', 'countryAR',
        'countryAM', 'countryAW', 'countryAU', 'countryAT', 'countryAZ',
        'countryBS', 'countryBH', 'countryBD', 'countryBB', 'countryBY',
        'countryBE', 'countryBZ', 'countryBJ', 'countryBM', 'countryBT',
        'countryBO', 'countryBA', 'countryBW', 'countryBV', 'countryBR',
        'countryIO', 'countryBN', 'countryBG', 'countryBF', 'countryBI',
        'countryKH', 'countryCM', 'countryCA', 'countryCV', 'countryKY',
        'countryCF', 'countryTD', 'countryCL', 'countryCN', 'countryCX',
        'countryCC', 'countryCO', 'countryKM', 'countryCG', 'countryCD',
        'countryCK', 'countryCR', 'countryCI', 'countryHR', 'countryCU',
        'countryCY', 'countryCZ', 'countryDK', 'countryDJ', 'countryDM',
        'countryDO', 'countryTP', 'countryEC', 'countryEG', 'countrySV',
        'countryGQ', 'countryER', 'countryEE', 'countryET', 'countryEU',
        'countryFK', 'countryFO', 'countryFJ', 'countryFI', 'countryFR',
        'countryFX', 'countryGF', 'countryPF', 'countryTF', 'countryGA',
        'countryGM', 'countryGE', 'countryDE', 'countryGH', 'countryGI',
        'countryGR', 'countryGL', 'countryGD', 'countryGP', 'countryGU',
        'countryGT', 'countryGN', 'countryGW', 'countryGY', 'countryHT',
        'countryHM', 'countryVA', 'countryHN', 'countryHK', 'countryHU',
        'countryIS', 'countryIN', 'countryID', 'countryIR', 'countryIQ',
        'countryIE', 'countryIL', 'countryIT', 'countryJM', 'countryJP',
        'countryJO', 'countryKZ', 'countryKE', 'countryKI', 'countryKP',
        'countryKR', 'countryKW', 'countryKG', 'countryLA', 'countryLV',
        'countryLB', 'countryLS', 'countryLR', 'countryLY', 'countryLI',
        'countryLT', 'countryLU', 'countryMO', 'countryMK', 'countryMG',
        'countryMW', 'countryMY', 'countryMV', 'countryML', 'countryMT',
        'countryMH', 'countryMQ', 'countryMR', 'countryMU', 'countryYT',
        'countryMX', 'countryFM', 'countryMD', 'countryMC', 'countryMN',
        'countryMS', 'countryMA', 'countryMZ', 'countryMM', 'countryNA',
        'countryNR', 'countryNP', 'countryNL', 'countryAN', 'countryNC',
        'countryNZ', 'countryNI', 'countryNE', 'countryNG', 'countryNU',
        'countryNF', 'countryMP', 'countryNO', 'countryOM', 'countryPK',
        'countryPW', 'countryPS', 'countryPA', 'countryPG', 'countryPY',
        'countryPE', 'countryPH', 'countryPN', 'countryPL', 'countryPT',
        'countryPR', 'countryQA', 'countryRE', 'countryRO', 'countryRU',
        'countryRW', 'countrySH', 'countryKN', 'countryLC', 'countryPM',
        'countryVC', 'countryWS', 'countrySM', 'countryST', 'countrySA',
        'countrySN', 'countryCS', 'countrySC', 'countrySL', 'countrySG',
        'countrySK', 'countrySI', 'countrySB', 'countrySO', 'countryZA',
        'countryGS', 'countryES', 'countryLK', 'countrySD', 'countrySR',
        'countrySJ', 'countrySZ', 'countrySE', 'countryCH', 'countrySY',
        'countryTW', 'countryTJ', 'countryTZ', 'countryTH', 'countryTG',
        'countryTK', 'countryTO', 'countryTT', 'countryTN', 'countryTR',
        'countryTM', 'countryTC', 'countryTV', 'countryUG', 'countryUA',
        'countryAE', 'countryUK', 'countryUS', 'countryUM', 'countryUY',
        'countryUZ', 'countryVU', 'countryVE', 'countryVN', 'countryVG',
        'countryVI', 'countryWF', 'countryEH', 'countryYE', 'countryYU',
        'countryZM', 'countryZW'
    },

    gl={
        'ad', 'ae', 'af', 'ag', 'ai', 'al', 'am', 'an', 'ao', 'aq', 'ar',
        'as', 'at', 'au', 'aw', 'az', 'ba', 'bb', 'bd', 'be', 'bf', 'bg',
        'bh', 'bi', 'bj', 'bm', 'bn', 'bo', 'br', 'bs', 'bt', 'bv', 'bw',
        'by', 'bz', 'ca', 'cc', 'cd', 'cf', 'cg', 'ch', 'ci', 'ck', 'cl',
        'cm', 'cn', 'co', 'cr', 'cs', 'cu', 'cv', 'cx', 'cy', 'cz', 'de',
        'dj', 'dk', 'dm', 'do', 'dz', 'ec', 'ee', 'eg', 'eh', 'er', 'es',
        'et', 'fi', 'fj', 'fk', 'fm', 'fo', 'fr', 'ga', 'gd', 'ge', 'gf',
        'gh', 'gi', 'gl', 'gm', 'gn', 'gp', 'gq', 'gr', 'gs', 'gt', 'gu',
        'gw', 'gy', 'hk', 'hm', 'hn', 'hr', 'ht', 'hu', 'id', 'ie', 'il',
        'in', 'io', 'iq', 'ir', 'is', 'it', 'jm', 'jo', 'jp', 'ke', 'kg',
        'kh', 'ki', 'km', 'kn', 'kp', 'kr', 'kw', 'ky', 'kz', 'la', 'lb',
        'lc', 'li', 'lk', 'lr', 'ls', 'lt', 'lu', 'lv', 'ly', 'ma', 'mc',
        'md', 'mg', 'mh', 'mk', 'ml', 'mm', 'mn', 'mo', 'mp', 'mq', 'mr',
        'ms', 'mt', 'mu', 'mv', 'mw', 'mx', 'my', 'mz', 'na', 'nc', 'ne',
        'nf', 'ng', 'ni', 'nl', 'no', 'np', 'nr', 'nu', 'nz', 'om', 'pa',
        'pe', 'pf', 'pg', 'ph', 'pk', 'pl', 'pm', 'pn', 'pr', 'ps', 'pt',
        'pw', 'py', 'qa', 're', 'ro', 'ru', 'rw', 'sa', 'sb', 'sc', 'sd',
        'se', 'sg', 'sh', 'si', 'sj', 'sk', 'sl', 'sm', 'sn', 'so', 'sr',
        'st', 'sv', 'sy', 'sz', 'tc', 'td', 'tf', 'tg', 'th', 'tj', 'tk',
        'tl', 'tm', 'tn', 'to', 'tr', 'tt', 'tv', 'tw', 'tz', 'ua', 'ug',
        'uk', 'um', 'us', 'uy', 'uz', 'va', 'vc', 've', 'vg', 'vi', 'vn',
        'vu', 'wf', 'ws', 'ye', 'yt', 'za', 'zm', 'zw',
    },

    filter={0, 1},

    hl={
        'af', 'sq', 'sm', 'ar', 'az', 'eu', 'be', 'bn', 'bh', 'bs', 'bg',
        'ca', 'zh-CN', 'zh-TW', 'hr', 'cs', 'da', 'nl', 'en', 'eo', 'et',
        'fo', 'fi', 'fr', 'fy', 'gl', 'ka', 'de', 'el', 'gu', 'iw', 'hi',
        'hu', 'is', 'id', 'ia', 'ga', 'it', 'ja', 'jw', 'kn', 'ko', 'la',
        'lv', 'lt', 'mk', 'ms', 'ml', 'mt', 'mr', 'ne', 'no', 'nn', 'oc',
        'fa', 'pl', 'pt-BR', 'pt-PT', 'pa', 'ro', 'ru', 'gd', 'sr', 'si',
        'sk', 'sl', 'es', 'su', 'sw', 'sv', 'tl', 'ta', 'te', 'th', 'ti',
        'tr', 'uk', 'ur', 'uz', 'vi', 'cy', 'xh', 'zu'
    },

    imgColorType={
        'color', 'gray', 'mono', 'trans'
    },

    imgDominantColor={
        'black',
        'blue',
        'brown',
        'gray',
        'green',
        'orange',
        'pink',
        'purple',
        'red',
        'teal',
        'white',
        'yellow',
    },

    imgSize={
        'huge',
        'icon',
        'large',
        'medium',
        'small',
        'xlarge',
        'xxlarge',
    },

    imgType={
        'clipart',
        'face',
        'lineart',
        'stock',
        'photo',
        'animated'
    },

    lr={
        'lang_ar', 'lang_bg', 'lang_ca', 'lang_zh-CN', 'lang_zh-TW',
        'lang_hr', 'lang_cs', 'lang_da', 'lang_nl', 'lang_en', 'lang_et',
        'lang_fi', 'lang_fr', 'lang_de', 'lang_el', 'lang_iw', 'lang_hu',
        'lang_is', 'lang_id', 'lang_it', 'lang_ja', 'lang_ko', 'lang_lv',
        'lang_lt', 'lang_no', 'lang_pl', 'lang_pt', 'lang_ro', 'lang_ru',
        'lang_sr', 'lang_sk', 'lang_sl', 'lang_es', 'lang_sv', 'lang_tr',
    },

    num={1, 2, 3, 4, 5, 6, 7, 8, 9, 10},

    rights={
        'cc_publicdomain', 'cc_attribute', 'cc_sharealike',
        'cc_noncommercial', 'cc_nonderived'
    },

    safe={'active', 'off'},

    searchType={None, 'image'},

    siteSearchFilter={'e', 'i'},

    start=range(1, 92)
)


##############################################################################
# YouTube variables
##############################################################################


YOUTUBE_TOPIC_IDS = {
    'Entertainment topics': {'Entertainment (parent topic)': '/m/02jjt',
                             'Humor': '/m/09kqc',
                             'Movies': '/m/02vxn',
                             'Performing arts': '/m/05qjc',
                             'Professional wrestling': '/m/066wd',
                             'TV shows': '/m/0f2f9'},
    'Gaming topics': {'Action game': '/m/025zzc',
                      'Action-adventure game': '/m/02ntfj',
                      'Casual game': '/m/0b1vjn',
                      'Gaming (parent topic)': '/m/0bzvm2',
                      'Music video game': '/m/02hygl',
                      'Puzzle video game': '/m/04q1x3q',
                      'Racing video game': '/m/01sjng',
                      'Role-playing video game': '/m/0403l3g',
                      'Simulation video game': '/m/021bp2',
                      'Sports game': '/m/022dc6',
                      'Strategy video game': '/m/03hf_rm'},
    'Lifestyle topics': {'Fashion': '/m/032tl',
                         'Fitness': '/m/027x7n',
                         'Food': '/m/02wbm',
                         'Hobby': '/m/03glg',
                         'Lifestyle (parent topic)': '/m/019_rr',
                         'Pets': '/m/068hy',
                         'Physical attractiveness [Beauty]': '/m/041xxh',
                         'Technology': '/m/07c1v',
                         'Tourism': '/m/07bxq',
                         'Vehicles': '/m/07yv9'},
    'Music topics': {'Christian music': '/m/02mscn',
                     'Classical music': '/m/0ggq0m',
                     'Country': '/m/01lyv',
                     'Electronic music': '/m/02lkt',
                     'Hip hop music': '/m/0glt670',
                     'Independent music': '/m/05rwpb',
                     'Jazz': '/m/03_d0',
                     'Music (parent topic)': '/m/04rlf',
                     'Music of Asia': '/m/028sqc',
                     'Music of Latin America': '/m/0g293',
                     'Pop music': '/m/064t9',
                     'Reggae': '/m/06cqb',
                     'Rhythm and blues': '/m/06j6l',
                     'Rock music': '/m/06by7',
                     'Soul music': '/m/0gywn'},
    'Other topics': {'Knowledge': '/m/01k8wb'},
    'Society topics': {'Business': '/m/09s1f',
                       'Health': '/m/0kt51',
                       'Military': '/m/01h6rj',
                       'Politics': '/m/05qt0',
                       'Religion': '/m/06bvp',
                       'Society (parent topic)': '/m/098wr'},
    'Sports topics': {'American football': '/m/0jm_',
                      'Baseball': '/m/018jz',
                      'Basketball': '/m/018w8',
                      'Boxing': '/m/01cgz',
                      'Cricket': '/m/09xp_',
                      'Football': '/m/02vx4',
                      'Golf': '/m/037hz',
                      'Ice hockey': '/m/03tmr',
                      'Mixed martial arts': '/m/01h7lh',
                      'Motorsport': '/m/0410tth',
                      'Sports (parent topic)': '/m/06ntj',
                      'Tennis': '/m/07bs0',
                      'Volleyball': '/m/07_53'}
                    }

YOUTUBE_VID_CATEGORY_IDS = {
    'Action/Adventure': '32',
    'Anime/Animation': '31',
    'Autos & Vehicles': '2',
    'Classics': '33',
    'Comedy': '34',
    'Documentary': '35',
    'Drama': '36',
    'Education': '27',
    'Entertainment': '24',
    'Family': '37',
    'Film & Animation': '1',
    'Foreign': '38',
    'Gaming': '20',
    'Horror': '39',
    'Howto & Style': '26',
    'Movies': '30',
    'Music': '10',
    'News & Politics': '25',
    'Nonprofits & Activism': '29',
    'People & Blogs': '22',
    'Pets & Animals': '15',
    'Sci-Fi/Fantasy': '40',
    'Science & Technology': '28',
    'Short Movies': '18',
    'Shorts': '42',
    'Shows': '43',
    'Sports': '17',
    'Thriller': '41',
    'Trailers': '44',
    'Travel & Events': '19',
    'Videoblogging': '21'
}

SERP_YTUBE_VALID_VALS = dict(
    channelType={'any', 'show'},

    eventType={'completed', 'live', 'upcoming'},

    forContentOwner={True, False, 'true', 'false'},

    forDeveloper={True, False, 'true', 'false'},

    forMine={True, False, 'true', 'false'},

    maxResults=range(51),

    order={'date', 'rating', 'relevance', 'title',
           'videoCount', 'viewCount'},

    regionCode={
        'ad', 'ae', 'af', 'ag', 'ai', 'al', 'am', 'an', 'ao', 'aq', 'ar',
        'as', 'at', 'au', 'aw', 'az', 'ba', 'bb', 'bd', 'be', 'bf', 'bg',
        'bh', 'bi', 'bj', 'bm', 'bn', 'bo', 'br', 'bs', 'bt', 'bv', 'bw',
        'by', 'bz', 'ca', 'cc', 'cd', 'cf', 'cg', 'ch', 'ci', 'ck', 'cl',
        'cm', 'cn', 'co', 'cr', 'cs', 'cu', 'cv', 'cx', 'cy', 'cz', 'de',
        'dj', 'dk', 'dm', 'do', 'dz', 'ec', 'ee', 'eg', 'eh', 'er', 'es',
        'et', 'fi', 'fj', 'fk', 'fm', 'fo', 'fr', 'ga', 'gd', 'ge', 'gf',
        'gh', 'gi', 'gl', 'gm', 'gn', 'gp', 'gq', 'gr', 'gs', 'gt', 'gu',
        'gw', 'gy', 'hk', 'hm', 'hn', 'hr', 'ht', 'hu', 'id', 'ie', 'il',
        'in', 'io', 'iq', 'ir', 'is', 'it', 'jm', 'jo', 'jp', 'ke', 'kg',
        'kh', 'ki', 'km', 'kn', 'kp', 'kr', 'kw', 'ky', 'kz', 'la', 'lb',
        'lc', 'li', 'lk', 'lr', 'ls', 'lt', 'lu', 'lv', 'ly', 'ma', 'mc',
        'md', 'mg', 'mh', 'mk', 'ml', 'mm', 'mn', 'mo', 'mp', 'mq', 'mr',
        'ms', 'mt', 'mu', 'mv', 'mw', 'mx', 'my', 'mz', 'na', 'nc', 'ne',
        'nf', 'ng', 'ni', 'nl', 'no', 'np', 'nr', 'nu', 'nz', 'om', 'pa',
        'pe', 'pf', 'pg', 'ph', 'pk', 'pl', 'pm', 'pn', 'pr', 'ps', 'pt',
        'pw', 'py', 'qa', 're', 'ro', 'ru', 'rw', 'sa', 'sb', 'sc', 'sd',
        'se', 'sg', 'sh', 'si', 'sj', 'sk', 'sl', 'sm', 'sn', 'so', 'sr',
        'st', 'sv', 'sy', 'sz', 'tc', 'td', 'tf', 'tg', 'th', 'tj', 'tk',
        'tl', 'tm', 'tn', 'to', 'tr', 'tt', 'tv', 'tw', 'tz', 'ua', 'ug',
        'uk', 'um', 'us', 'uy', 'uz', 'va', 'vc', 've', 'vg', 'vi', 'vn',
        'vu', 'wf', 'ws', 'ye', 'yt', 'za', 'zm', 'zw',
    },

    relevanceLanguage={
        'af', 'sq', 'sm', 'ar', 'az', 'eu', 'be', 'bn', 'bh', 'bs', 'bg',
        'ca', 'zh-CN', 'zh-TW', 'zh-Hans', 'zh-Hant', 'hr', 'cs', 'da',
        'nl', 'en', 'eo', 'et', 'fo', 'fi', 'fr', 'fy', 'gl', 'ka', 'de',
        'el', 'gu', 'iw', 'hi', 'hu', 'is', 'id', 'ia', 'ga', 'it', 'ja',
        'jw', 'kn', 'ko', 'la', 'lv', 'lt', 'mk', 'ms', 'ml', 'mt', 'mr',
        'ne', 'no', 'nn', 'oc', 'fa', 'pl', 'pt-BR', 'pt-PT', 'pa', 'ro',
        'ru', 'gd', 'sr', 'si', 'sk', 'sl', 'es', 'su', 'sw', 'sv', 'tl',
        'ta', 'te', 'th', 'ti', 'tr', 'uk', 'ur', 'uz', 'vi', 'cy', 'xh',
        'zu'
    },

    safeSearch={'moderate', 'none', 'strict'},

    topicId={
        '/m/04rlf', '/m/02mscn', '/m/0ggq0m', '/m/01lyv', '/m/02lkt',
        '/m/0glt670', '/m/05rwpb', '/m/03_d0', '/m/028sqc', '/m/0g293',
        '/m/064t9', '/m/06cqb', '/m/06j6l', '/m/06by7', '/m/0gywn',
        '/m/0bzvm2', '/m/025zzc', '/m/02ntfj', '/m/0b1vjn', '/m/02hygl',
        '/m/04q1x3q', '/m/01sjng', '/m/0403l3g', '/m/021bp2', '/m/022dc6',
        '/m/03hf_rm', '/m/06ntj', '/m/0jm_', '/m/018jz', '/m/018w8',
        '/m/01cgz', '/m/09xp_', '/m/02vx4', '/m/037hz', '/m/03tmr',
        '/m/01h7lh', '/m/0410tth', '/m/07bs0', '/m/07_53', '/m/02jjt',
        '/m/09kqc', '/m/02vxn', '/m/05qjc', '/m/066wd', '/m/0f2f9',
        '/m/019_rr', '/m/032tl', '/m/027x7n', '/m/02wbm', '/m/03glg',
        '/m/068hy', '/m/041xxh', '/m/07c1v', '/m/07bxq', '/m/07yv9',
        '/m/098wr', '/m/09s1f', '/m/0kt51', '/m/01h6rj', '/m/05qt0',
        '/m/06bvp', '/m/01k8wb'
    },

    type={'channel', 'playlist', 'video'},

    videoCaption={'any', 'closedCaption', 'none'},

    videoCategoryId={
        '1', '2', '10', '15', '17', '18', '19', '20', '21', '22', '23',
        '24', '25', '26', '27', '28', '29', '30', '31', '32', '33', '34',
        '35', '36', '37', '38', '39', '40', '41', '42', '43', '44'
    },

    videoDefinition={'any', 'high', 'standard'},

    videoDimension={'2d', '3d', 'any'},

    videoDuration={'any', 'long', 'medium', 'short'},

    videoEmbeddable={'any', True, 'true'},

    videoLicense={'any', 'creativeCommon', 'youtube'},

    videoSyndicated={'any', True, 'true'},

    videoType={'any', 'episode', 'movie'},
)


def _split_by_comma(s, length=50):
    """Group a comma-separated string into a list of at-most
    ``length``-length words each."""
    str_split = s.split(',')
    str_list = []
    for i in range(0, len(str_split) + length, length):
        temp_str = ','.join(str_split[i:i+length])
        if temp_str:
            str_list.append(temp_str)
    return str_list


def youtube_video_details(key, vid_ids):
    """Return details of videos for which the ids are given.
    Assumes ``ids`` is a comma-separated list of video ids with
    no spaces."""
    base_url = ('https://www.googleapis.com/youtube/v3/videos?part='
                'contentDetails,id,liveStreamingDetails,localizations,player,'
                'recordingDetails,snippet,statistics,status,topicDetails')
    vid_ids = _split_by_comma(vid_ids, length=50)
    final_df = pd.DataFrame()
    for vid_id in vid_ids:
        params = {'id': vid_id, 'key': key}
        logging.info(msg='Requesting: ' + 'video details')
        video_resp = requests.get(base_url, params=params)
        if video_resp.status_code >= 400:
            raise Exception(video_resp.json())
        items_df = pd.DataFrame(video_resp.json()['items'])
        details = ['snippet', 'topicDetails', 'statistics',
                   'status', 'contentDetails']
        detail_df = pd.DataFrame()
        for detail in details:
            try:
                detail_df = pd.concat([
                    detail_df,
                    pd.DataFrame([x[detail] for x in
                                  video_resp.json()['items']])
                ], axis=1)
            except KeyError:
                continue
        temp_df = pd.concat([items_df, detail_df], axis=1)
        final_df = pd.concat([final_df, temp_df], sort=False, ignore_index=True)
    return final_df


def youtube_channel_details(key, channel_ids):
    """Return details of channels for which the ids are given.
    Assumes ``ids`` is a comma-separated list of channel ids with
    no spaces."""
    base_url = ('https://www.googleapis.com/youtube/v3/channels?part='
                'snippet,contentDetails,statistics')
    channel_ids = _split_by_comma(channel_ids, length=50)
    final_df = pd.DataFrame()
    for channel_id in channel_ids:
        params = {'id': channel_id, 'key': key}
        logging.info(msg='Requesting: ' + 'channel details')
        channel_resp = requests.get(base_url, params=params)
        if channel_resp.status_code >= 400:
            raise Exception(channel_resp.json())
        items_df = pd.DataFrame(channel_resp.json()['items'])
        details = ['snippet', 'statistics', 'contentDetails']
        detail_df = pd.DataFrame()
        for detail in details:
            try:
                detail_df = pd.concat([
                    detail_df,
                    pd.DataFrame([x[detail] for x in
                                  channel_resp.json()['items']])
                ], axis=1)
            except KeyError:
                continue
        temp_df = pd.concat([items_df, detail_df], axis=1)
        final_df = pd.concat([final_df, temp_df], sort=False, ignore_index=True)
    return final_df


def _dict_product(d):
    """Return the product of all values of a dict, while
    coupling each value with its key.
    This is used to generate multiple queries out of
        possibly multiple arguments in serp_goog.

    >>> d = {'a': [1], 'b': [2, 3, 4], 'c': [5, 6]}
    >>> _dict_product(d)
    >>> [{'a': 1, 'b': 2, 'c': 5},
         {'a': 1, 'b': 2, 'c': 6},
         {'a': 1, 'b': 3, 'c': 5},
         {'a': 1, 'b': 3, 'c': 6},
         {'a': 1, 'b': 4, 'c': 5},
         {'a': 1, 'b': 4, 'c': 6}]
    """
    items = list(d.items())
    keys = [x[0] for x in items]
    values = [x[1] for x in items]
    dicts = []
    for prod in product(*values):
        tempdict = dict(zip(keys, prod))
        dicts.append(tempdict)
    return dicts


def serp_goog(q, cx, key, c2coff=None, cr=None,
              dateRestrict=None, exactTerms=None, excludeTerms=None,
              fileType=None, filter=None, gl=None, highRange=None,
              hl=None, hq=None, imgColorType=None, imgDominantColor=None,
              imgSize=None, imgType=None, linkSite=None, lowRange=None,
              lr=None, num=None, orTerms=None, relatedSite=None,
              rights=None, safe=None, searchType=None, siteSearch=None,
              siteSearchFilter=None, sort=None, start=None):
    """Query Google and get search results in a DataFrame.

    For each parameter, you can supply single or multiple values / arguments.
        If you pass multiple arguments, all the possible combinations of
        arguments (the product) will be requested, and you will get one
        DataFrame combining all queries. See examples below.

    :param q: The search expression.
    :param cx: The custom search engine ID to use for this
        request.
    :param key: The API key of your custom search engine.
    :param c2coff: Enables or disables Simplified and
        Traditional Chinese Search. The default value for this
        parameter is 0 (zero), meaning that the feature is enabled.
        Supported values are:1: Disabled0: Enabled (default)
    :param cr: Restricts search results to documents
        originating in a particular country. You may use Boolean
        operators in the cr parameter's value.Google Search
        determines the country of a document by analyzing:the top-
        level domain (TLD) of the document's URLthe geographic
        location of the Web server's IP addressSee the Country
        Parameter Values page for a list of valid values for this
        parameter.
    :param dateRestrict: Restricts results to URLs based on
        date. Supported values include:d[number]: requests results
        from the specified number of past days.
        - d[number]: requests results from the specified number of past days.
        - w[number]: requests results from the specified number of past weeks.
        - m[number]: requests results from the specified number of past months.
        - y[number]: requests results from the specified number of past years.
    :param exactTerms: Identifies a phrase that all
        documents in the search results must contain.
    :param excludeTerms: Identifies a word or phrase that
        should not appear in any documents in the search results.
    :param fileType: Restricts results to files of a
        specified extension. A list of file types indexable by
        Google can be found in Search Console Help Center.
    :param filter: Controls turning on or off the duplicate
        content filter.See Automatic Filtering for more information
        about Google's search results filters. Note that host
        crowding filtering applies only to multi-site searches.By
        default, Google applies filtering to all search results to
        improve the quality of those results.  Acceptable values
        are:  "0": Turns off duplicate content filter.  "1": Turns
        on duplicate content filter.
    :param gl: Geolocation of end user. The gl parameter
        value is a two-letter country code. The gl parameter boosts
        search results whose country of origin matches the parameter
        value. See the Country Codes page for a list of valid
        values.Specifying a gl parameter value should lead to more
        relevant results. This is particularly true for
        international customers and, even more specifically, for
        customers in English- speaking countries other than the
        United States.
    :param highRange: Specifies the ending value for a
        search range.Use lowRange and highRange to append an
        inclusive search range of lowRange...highRange to the query.
    :param hl: Sets the user interface language. Explicitly
        setting this parameter improves the performance and the
        quality of your search results.See the Interface
        Languages section of Internationalizing Queries and Results
        Presentation for more information, and Supported Interface
        Languages for a list of supported languages.
    :param hq: Appends the specified query terms to the
        query, as if they were combined with a logical AND operator.
    :param imgColorType: Returns black and white, grayscale,
        or color images: mono, gray, and color.  Acceptable values
        are:  "color": color  "gray": gray  "mono": mono
    :param imgDominantColor: Returns images of a specific
        dominant color.  Acceptable values are:  "black": black
        "blue": blue  "brown": brown  "gray": gray  "green": green
        "orange": orange  "pink": pink  "purple": purple  "red": red
        "teal": teal  "white": white  "yellow": yellow
    :param imgSize: Returns images of a specified size.
        Acceptable values are:  "huge": huge  "icon": icon  "large":
        large  "medium": medium  "small": small  "xlarge": xlarge
        "xxlarge": xxlarge
    :param imgType: Returns images of a type.  Acceptable
        values are:  "clipart": clipart  "face": face  "lineart":
        lineart  "news": news  "photo": photo
    :param linkSite: Specifies that all search results
        should contain a link to a particular URL
    :param lowRange: Specifies the starting value for a
        search range. Use lowRange and highRange to append an
        inclusive search range of lowRange...highRange to the query.
    :param lr: Restricts the search to documents written in
        a particular language (e.g., lr=lang_ja).  Acceptable values
        are:  "lang_ar": Arabic  "lang_bg": Bulgarian  "lang_ca":
        Catalan  "lang_cs": Czech  "lang_da": Danish  "lang_de":
        German  "lang_el": Greek  "lang_en": English  "lang_es":
        Spanish  "lang_et": Estonian  "lang_fi": Finnish  "lang_fr":
        French  "lang_hr": Croatian  "lang_hu": Hungarian
        "lang_id": Indonesian  "lang_is": Icelandic  "lang_it":
        Italian  "lang_iw": Hebrew  "lang_ja": Japanese  "lang_ko":
        Korean  "lang_lt": Lithuanian  "lang_lv": Latvian
        "lang_nl": Dutch  "lang_no": Norwegian  "lang_pl": Polish
        "lang_pt": Portuguese  "lang_ro": Romanian  "lang_ru":
        Russian  "lang_sk": Slovak  "lang_sl": Slovenian  "lang_sr":
        Serbian  "lang_sv": Swedish  "lang_tr": Turkish  "lang_zh-
        CN": Chinese (Simplified)  "lang_zh-TW": Chinese
        (Traditional)
    :param num: Number of search results to return.Valid
        values are integers between 1 and 10, inclusive.
    :param orTerms: Provides additional search terms to
        check for in a document, where each document in the search
        results must contain at least one of the additional search
        terms.
    :param relatedSite: Specifies that all search results
        should be pages that are related to the specified URL.
    :param rights: Filters based on licensing. Supported
        values include: cc_publicdomain, cc_attribute,
        cc_sharealike, cc_noncommercial, cc_nonderived, and
        combinations of these.
    :param safe: Search safety level.  Acceptable values
        are:  "active": Enables SafeSearch filtering.  "off":
        Disables SafeSearch filtering.  (default)
    :param searchType: Specifies the search type: image. If
        unspecified, results are limited to webpages.  Acceptable
        values are:  "image": custom image search.
    :param siteSearch: Specifies all search results should
        be pages from a given site.
    :param siteSearchFilter: Controls whether to include or
        exclude results from the site named in the siteSearch
        parameter.  Acceptable values are:  "e": exclude  "i":
        include
    :param sort: The sort expression to apply to the
        results.
    :param start: The index of the first result to
        return.Valid value are integers starting 1 (default) and the
        second result is 2 and so forth. For example &start=11 gives
        the second page of results with the default "num" value of
        10 results per page.Note: No more than 100 results will ever
        be returned for any query with JSON API, even if more than
        100 documents match the query, so setting (start + num) to
        more than 100 will produce an error. Note that the maximum
        value for num is 10.

    The following function call will produce two queries:
    "hotel" in the USA, and "hotel" in France

    >>> serp_goog(q='hotel', gl=['us', 'fr'], cx='YOUR_CX', key='YOUR_KEY')

    The below function call will prouce four queries and make four requests:

    * "fligts" in UK
    * "fligts" in Australia
    * "tickets" in UK
    * "tickets" in Australia

    'cr' here refers to 'country restrict', which focuses on content
    originating from the specified country.

    >>> serp_goog(q=['flights', 'tickets'], cr=['countryUK', 'countryAU'],
                  cx='YOUR_CX', key='YOUR_KEY')
    """
    params = locals()
    supplied_params = {k: v for k, v in params.items() if params[k] is not None}

    for p in supplied_params:
        if isinstance(supplied_params[p], (str, int)):
            supplied_params[p] = [supplied_params[p]]

    for p in supplied_params:
        if p in SERP_GOOG_VALID_VALS:
            if not set(supplied_params[p]).issubset(SERP_GOOG_VALID_VALS[p]):
                raise ValueError('Please make sure you provide a'
                                 ' valid value for "{}", valid values:\n'
                                 '{}'.format(p,
                                             sorted(SERP_GOOG_VALID_VALS[p])))
    params_list = _dict_product(supplied_params)
    base_url = 'https://www.googleapis.com/customsearch/v1?'
    specified_cols = ['searchTerms', 'rank', 'title', 'snippet',
                      'displayLink', 'link', 'queryTime', 'totalResults']
    responses = []
    for param in params_list:
        param_log = ', '.join([k + '=' + str(v) for k, v in param.items()])
        logging.info(msg='Requesting: ' + param_log)
        resp = requests.get(base_url, params=param)
        if resp.status_code >= 400:
            raise Exception(resp.json())
        responses.append(resp)
    result_df = pd.DataFrame()
    for i, resp in enumerate(responses):
        request_metadata = resp.json()['queries']['request'][0]
        del request_metadata['title']
        search_info = resp.json()['searchInformation']
        if int(search_info['totalResults']) == 0:
            df = pd.DataFrame(columns=specified_cols, index=range(1))
            df['searchTerms'] = request_metadata['searchTerms']
            # These keys don't appear in the response so they have to be
            # added manually
            for missing in ['lr', 'num', 'start', 'c2coff']:
                if missing in params_list[i]:
                    df[missing] = params_list[i][missing]
        else:
            df = pd.DataFrame(resp.json()['items'])
            df['cseName'] = resp.json()['context']['title']
            start_idx = request_metadata['startIndex']
            df['rank'] = range(start_idx, start_idx + len(df))
            for missing in ['lr', 'num', 'start', 'c2coff']:
                if missing in params_list[i]:
                    df[missing] = params_list[i][missing]
        meta_columns = {**request_metadata, **search_info}
        df = df.assign(**meta_columns)
        df['queryTime'] = datetime.datetime.now(tz=datetime.timezone.utc)
        df['queryTime'] = pd.to_datetime(df['queryTime'])
        if 'image' in df:
            img_df = json_normalize(df['image'])
            img_df.columns = ['image.' + c for c in img_df.columns]
            df = pd.concat([df, img_df], axis=1)
        result_df = pd.concat([result_df, df], sort=False, ignore_index=True)
    ordered_cols = (list(set(params_list[i]).difference({'q', 'key', 'cx'})) +
                    specified_cols)
    non_ordered = result_df.columns.difference(set(ordered_cols))
    final_df = result_df[ordered_cols + list(non_ordered)]
    if 'pagemap' in final_df:
        pagemap_df = pd.DataFrame()
        for p in final_df['pagemap']:
            try:
                temp_pagemap_df = json_normalize(p)
                pagemap_df = pd.concat([pagemap_df, temp_pagemap_df], sort=False)
            except Exception as e:
                temp_pagemap_df = pd.DataFrame({'delete_me': None},
                                                index=range(1))
                pagemap_df = pd.concat([pagemap_df, temp_pagemap_df], sort=False)
        pagemap_df = pagemap_df.reset_index(drop=True)
        if 'delete_me' in pagemap_df:
            del pagemap_df['delete_me']
        for col in pagemap_df:
            if col in final_df:
                pagemap_df = pagemap_df.rename(columns={col: 'pagemap_' + col})
        final_df = pd.concat([final_df, pagemap_df], axis=1)

        if 'metatags' in pagemap_df:
            metatag_df = pd.DataFrame()
            for m in pagemap_df['metatags']:
                try:
                    temp_metatags_df = json_normalize(m)
                    metatag_df = pd.concat([metatag_df, temp_metatags_df],
                                           sort=False)
                except Exception as e:
                    temp_metatags_df = pd.DataFrame({'delete_me': None},
                                                    index=range(1))
                    metatag_df = pd.concat([metatag_df, temp_metatags_df],
                                           sort=False)
            metatag_df = metatag_df.reset_index(drop=True)
            if 'delete_me' in metatag_df:
                del metatag_df['delete_me']
            for col in metatag_df:
                if col in final_df:
                    metatag_df = metatag_df.rename(columns={col: 'metatag_' + col})

            final_df = pd.concat([final_df, metatag_df], axis=1)
    return final_df


def serp_youtube(key, q=None, channelId=None, channelType=None, eventType=None,
                 forContentOwner=None, forDeveloper=None, forMine=None,
                 location=None, locationRadius=None, maxResults=None,
                 onBehalfOfContentOwner=None, order=None, pageToken=None,
                 publishedAfter=None, publishedBefore=None, regionCode=None,
                 relatedToVideoId=None, relevanceLanguage=None,
                 safeSearch=None, topicId=None, type=None, videoCaption=None,
                 videoCategoryId=None, videoDefinition=None,
                 videoDimension=None, videoDuration=None, videoEmbeddable=None,
                 videoLicense=None, videoSyndicated=None, videoType=None):
    """Query the YouTube API and get search results in a DataFrame.
    For each parameter you can supply a single or multiple value(s).
    Looping and merging results is handled automatically in case of multiple
    values.

    :param q: (string) The ``q`` parameter specifies the query term to
        search for. Your request can also use the Boolean NOT (-) and OR (|)
        operators to exclude videos or to find videos that are associated with
        one of several search terms. For example, to search for videos
        matching either "boating" or "sailing", set the ``q`` parameter value
        to boating|sailing. Similarly, to search for videos matching either
        "boating" or "sailing" but not "fishing", set the q parameter value to
        boating|sailing -fishing. Note that the pipe character must be URL-
        escaped when it is sent in your API request. The URL-escaped value for
        the pipe character is %7C.
    :param channelId: (string) The ``channelId`` parameter
        indicates that the API response should only contain resources
        created by the channel. Note: Search results are constrained
        to a maximum of 500 videos if your request specifies a value
        for the ``channelId`` parameter and sets the ``type`` parameter value
        to video, but it does not also set one of the ``forContentOwner``,
        ``forDeveloper``, or ``forMine`` filters.
    :param channelType: (string) The ``channelType`` parameter
        lets you restrict a search to a particular type of
        channel. Acceptable values are:

        any – Return all channels.

        show – Only retrieve shows.
    :param eventType: (string) The ``eventType`` parameter
        restricts a search to broadcast events. If you specify a value
        for this parameter, you must also set the type parameter's
        value to video. Acceptable values are:

        completed – Only include completed broadcasts.

        live – Only include active broadcasts.

        upcoming – Only include upcoming broadcasts.

    :param forContentOwner: (boolean) This parameter can
        only be used in a properly authorized request, and it is
        intended exclusively for YouTube content partners. The
        ``forContentOwner`` parameter restricts the search to only
        retrieve videos owned by the content owner identified by
        the ``onBehalfOfContentOwner`` parameter. If ``forContentOwner``
        is set to true, the request must also meet these
        requirements:The ``onBehalfOfContentOwner`` parameter is
        required.The user authorizing the request must be using
        an account linked to the specified content owner. The
        ``type`` parameter value must be set to video.None of the
        following other parameters can be set: ``videoDefinition``,
        ``videoDimension``, ``videoDuration``, ``videoLicense``,
        ``videoEmbeddable``, ``videoSyndicated``, ``videoType``.
    :param forDeveloper: (boolean) This parameter can only
        be used in a properly authorized request. The ``forDeveloper``
        parameter restricts the search to only retrieve videos
        uploaded via the developer's application or website. The
        API server uses the request's authorization credentials to
        identify the developer. The ``forDeveloper`` parameter can be
        used in conjunction with optional search parameters like
        the ``q`` parameter. For this feature, each uploaded video is
        automatically tagged with the project number that is
        associated with the developer's application in the Google
        Developers Console. When a search request subsequently sets
        the ``forDeveloper`` parameter to ``true`` the API server uses the
        request's authorization credentials to identify the
        developer. Therefore, a developer can restrict results to
        videos uploaded through the developer's own app or website
        but not to videos uploaded through other apps or sites.
    :param forMine: (boolean) This parameter can only be used in
        a properly authorized request. The ``forMine`` parameter restricts
        the search to only retrieve videos owned by the authenticated
        user. If you set this parameter to ``true``, then the ``type``
        parameter's value must also be set to ``video``. In addition, none
        of the following other parameters can be set in the same
        request: ``videoDefinition``, ``videoDimension``, ``videoDuration``,
        ``videoLicense``, ``videoEmbeddable``, ``videoSyndicated``,
        ``videoType``.
    :param relatedToVideoId: (string) The
        ``relatedToVideoId`` parameter retrieves a list of videos
        that are related to the video that the parameter ``value``
        identifies. The parameter ``value`` must be set to a
        YouTube video ID and, if you are using this parameter,
        the ``type`` parameter must be set to video.Note that if
        the ``relatedToVideoId`` parameter is set, the only other
        supported parameters are ``part``, ``maxResults``, ``pageToken``,
        ``regionCode``, ``relevanceLanguage``, ``safeSearch``, ``type`` (which
        must be set to video), and ``fields``.
    :param location: (string) The ``location`` parameter, in
        conjunction with the ``locationRadius`` parameter, defines a
        circular geographic area and also restricts a search to videos
        that specify, in their metadata, a geographic location that
        falls within that area. The parameter value is a string that
        specifies latitude/longitude coordinates e.g.
        (37.42307,-122.08427).The location parameter value identifies
        the point at the center of the area. The ``locationRadius``
        parameter specifies the maximum distance that the location
        associated with a video can be from that point for the video to
        still be included in the search results. The API returns an
        error if your request specifies a value for the ``location``
        parameter but does not also specify a value for the
        ``locationRadius`` parameter.
    :param locationRadius: (string) The ``locationRadius``
        parameter, in conjunction with the ``location`` parameter,
        defines a circular geographic area. The parameter value
        must be a floating point number followed by a measurement
        unit. Valid measurement units are m, km, ft, and mi. For
        example, valid parameter values include 1500m, 5km,
        10000ft, and 0.75mi. The API does not support
        ``locationRadius`` parameter values larger than 1000
        kilometers. Note: See the definition of the ``location``
        parameter for more information.
    :param maxResults: (unsigned integer) The ``maxResults``
        parameter specifies the maximum number of items that should
        be returned in the result set. Acceptable values are 0 to 50,
        inclusive. The default value is 5.
    :param onBehalfOfContentOwner: (string) This
        parameter can only be used in a properly
        authorized request. Note: This parameter is
        intended exclusively for YouTube content
        partners.The ``onBehalfOfContentOwner`` parameter
        indicates that the request's authorization
        credentials identify a YouTube CMS user who is
        acting on behalf of the content owner specified
        in the parameter value. This parameter is
        intended for YouTube content partners that own
        and manage many different YouTube channels. It
        allows content owners to authenticate once and
        get access to all their video and channel data,
        without having to provide authentication
        credentials for each individual channel. The CMS
        account that the user authenticates with must be
        linked to the specified YouTube content owner.
    :param order: (string) The order parameter specifies the
        method that will be used to order resources in the API response.
        The default value is relevance. Acceptable values are:

        date – Resources are sorted in reverse chronological order based on the
        date they were created.

        rating – Resources are sorted from highest to lowest rating.

        relevance – Resources are sorted based on their relevance to the search
        query. This is the default value for this parameter.

        title – Resources are sorted alphabetically by title.

        videoCount – Channels are sorted in descending order of their number of
        uploaded videos.

        viewCount – Resources sorted from highest to lowest number of views.
        For live broadcasts, videos are sorted by number of concurrent viewers
        while the broadcasts are ongoing.
    :param pageToken: (string) The ``pageToken`` parameter
        identifies a specific page in the result set that should be
        returned. In an API response, the ``nextPageToken`` and
        ``prevPageToken`` properties identify other pages that could be
        retrieved.
    :param publishedAfter: (datetime) The ``publishedAfter``
        parameter indicates that the API response should only
        contain resources created at or after the specified time.
        The value is an RFC 3339 formatted date-time value
        (1970-01-01T00:00:00Z).
    :param publishedBefore: (datetime) The
        ``publishedBefore`` parameter indicates that the API
        response should only contain resources created before or
        at the specified time. The value is an RFC 3339
        formatted date-time value (1970-01-01T00:00:00Z).
    :param regionCode: (string) The ``regionCode`` parameter
        instructs the API to return search results for videos that
        can be viewed in the specified country. The parameter value
        is an ISO 3166-1 alpha-2 country code.
    :param relevanceLanguage: (string) The
        ``relevanceLanguage`` parameter instructs the API to
        return search results that are most relevant to the
        specified language. The parameter value is typically
        an ISO 639-1 two-letter language code. However, you
        should use the values zh-Hans for simplified Chinese
        and zh-Hant for traditional Chinese. Please note that
        results in other languages will still be returned if
        they are highly relevant to the search query term.
    :param safeSearch: (string) The ``safeSearch`` parameter
        indicates whether the search results should include
        restricted content as well as standard content. Acceptable
        values are:

        moderate – YouTube will filter some content from search results and,
        at the least, will filter content that is restricted in your locale.
        Based on their content, search results could be removed from search
        results or demoted in search results. This is the default parameter
        value.

        none – YouTube will not filter the search result set.

        strict – YouTube will try to exclude all restricted content from the
        search result set.

        Based on their content, search results
        could be removed from search results or demoted in search
        results.
    :param topicId: (string) The ``topicId`` parameter indicates
        that the API response should only contain resources associated
        with the specified topic. The value identifies a Freebase topic
        ID.
    :param type: (string) The ``type`` parameter restricts a search
        query to only retrieve a particular type of resource. The value is
        a comma-separated list of resource types. The default value is
        video,channel,playlist. Acceptable values are: channel, playlist, and
        video
    :param videoCaption: (string) The ``videoCaption``
        parameter indicates whether the API should filter video
        search results based on whether they have captions. If you
        specify a value for this parameter, you must also set the
        ``type`` parameter's value to video. Acceptable values are:

        any – Do not filter results based on caption availability.

        closedCaption – Only include videos that have captions.

        none – Only include videos that do not have captions.
    :param videoCategoryId: (string) The ``videoCategoryId``
        parameter filters video search results based on their
        category. If you specify a value for this parameter, you
        must also set the ``type`` parameter's value to video.
    :param videoDefinition: (string) The ``videoDefinition``
        parameter lets you restrict a search to only include
        either high definition (HD) or standard definition (SD)
        videos. HD videos are available for playback in at least
        720p, though higher resolutions, like 1080p, might also
        be available. If you specify a value for this parameter,
        you must also set the ``type`` parameter's value to
        video. Acceptable values are:

        any – Return all videos, regardless of their resolution.

        high – Only retrieve HD videos.

        standard – Only retrieve videos in standard definition.
    :param videoDimension: (string) The ``videoDimension``
        parameter lets you restrict a search to only retrieve 2D
        or 3D videos. If you specify a value for this parameter,
        you must also set the ``type`` parameter's value to
        video. Acceptable values are:

        2d – Restrict search results to exclude 3D videos.

        3d – Restrict search results to only include 3D videos.

        any – Include both 3D and non-3D videos in returned results.
        This is the default value.
    :param videoDuration: (string) The ``videoDuration``
        parameter filters video search results based on their
        duration. If you specify a value for this parameter, you
        must also set the ``type`` parameter's value to
        video. Acceptable values are:

        any – Do not filter video search results based on their duration.
        This is the default value.

        long – Only include videos longer than 20 minutes.

        medium – Only include videos that are between four and 20 minutes
        long (inclusive).

        short – Only include videos that are less than four minutes long.
    :param videoEmbeddable: (string) The ``videoEmbeddable``
        parameter lets you to restrict a search to only videos
        that can be embedded into a webpage. If you specify a
        value for this parameter, you must also set the ``type``
        parameter's value to video. Acceptable values are:

        any – Return all videos, embeddable or not.

        true – Only retrieve embeddable videos.
    :param videoLicense: (string) The ``videoLicense``
        parameter filters search results to only include videos
        with a particular license. YouTube lets video uploaders
        choose to attach either the Creative Commons license or the
        standard YouTube license to each of their videos. If you
        specify a value for this parameter, you must also set the
        ``type`` parameter's value to video. Acceptable values are:

        any – Return all videos, regardless of which license they have,
        that match the query parameters.

        creativeCommon – Only return videos that have a Creative Commons
        license.
        Users can reuse videos with this license in other videos that they
        create.

        youtube – Only return videos that have the standard YouTube license.
    :param videoSyndicated: (string) The ``videoSyndicated``
        parameter lets you to restrict a search to only videos
        that can be played outside youtube.com. If you specify a
        value for this parameter, you must also set the ``type``
        parameter's value to video. Acceptable values are:

        any – Return all videos, syndicated or not.

        true – Only retrieve syndicated videos.
    :param videoType: (string) The ``videoType`` parameter lets
        you restrict a search to a particular type of videos. If you
        specify a value for this parameter, you must also set the ``type``
        parameter's value to video. Acceptable values are:

        any – Return all videos.

        episode – Only retrieve episodes of shows.

        movie – Only retrieve movies.
    """
    params = locals()
    supplied_params = {k: v for k, v in params.items() if params[k]}

    type_vid_params = {'eventType', 'relatedToVideoId', 'videoCaption',
                       'videoCategoryId', 'videoDefinition', 'videoDimension',
                       'videoDuration', 'videoEmbeddable', 'videoLicense',
                       'videoSyndicated', 'videoType', 'forMine',
                       'forContentOwner'}

    if (supplied_params.get('type') != 'video' and
            type_vid_params.intersection(set(supplied_params.keys()))):
        raise ValueError('You need to set type="video" if you want to set'
                         ' any of the following:' + str(type_vid_params))

    for p in supplied_params:
        if isinstance(supplied_params[p], (str, int)):
            supplied_params[p] = [supplied_params[p]]

    for p in supplied_params:
        if p in SERP_YTUBE_VALID_VALS:
            if not set(supplied_params[p]).issubset(SERP_YTUBE_VALID_VALS[p]):
                raise ValueError('Please make sure you provide a'
                                 ' valid value for "{}", valid values:\n{}'
                                 .format(p,
                                         sorted([str(x) for x in
                                                 SERP_YTUBE_VALID_VALS[p]])))

    params_list = _dict_product(supplied_params)
    base_url = "https://www.googleapis.com/youtube/v3/search?part=snippet"

    responses = []
    for param in params_list:
        param_log = ', '.join([k + '=' + str(v) for k, v in param.items()])
        logging.info(msg='Requesting: ' + param_log)
        resp = requests.get(base_url, params=param)
        if resp.status_code >= 400:
            raise Exception(resp.json())
        responses.append(resp)

    result_df = pd.DataFrame()
    for i, resp in enumerate(responses):
        snippet_df = pd.DataFrame([x['snippet'] for x in resp.json()['items']])
        id_df = pd.DataFrame([x['id'] for x in resp.json()['items']])
        if 'channelId' in id_df:
            id_df = id_df.drop('channelId', axis=1)

        if 'thumbnails' in snippet_df:
            thumb_df = json_normalize(snippet_df['thumbnails'])
        else:
            thumb_df = pd.DataFrame()
        page_info = resp.json()['pageInfo']
        temp_df = pd.concat([snippet_df, id_df, thumb_df],
                            axis=1).assign(**page_info)
        temp_df['rank'] = range(1, len(temp_df)+1)

        if len(temp_df) == 0:
            empty_df_cols = ['title', 'description', 'publishedAt',
                             'channelTitle', 'kind', 'videoId', 'channelId']
            temp_df = temp_df.assign(q=[params_list[i]['q']])
            temp_df = temp_df.assign(**dict.fromkeys(empty_df_cols))

            temp_df = temp_df.assign(**page_info)
        del params_list[i]['key']
        temp_df = temp_df.assign(**params_list[i])
        temp_df['nextPageToken'] = resp.json().get('nextPageToken')
        result_df = pd.concat([result_df, temp_df], sort=False,
                              ignore_index=True)

    result_df['queryTime'] = datetime.datetime.now(tz=datetime.timezone.utc)
    result_df['queryTime'] = pd.to_datetime(result_df['queryTime'])

    specified_cols = ['queryTime', 'rank', 'title', 'description',
                      'publishedAt', 'channelTitle', 'totalResults',
                      'kind']
    ordered_cols = list(params_list[i].keys()) + specified_cols
    non_ordered = result_df.columns.difference(set(ordered_cols))
    final_df = result_df[ordered_cols + list(non_ordered)]

    vid_ids = ','.join(final_df['videoId'].dropna())
    if vid_ids:
        vid_details_df = youtube_video_details(vid_ids=vid_ids, key=key)
        vid_details_df.columns = ['video.' + x for x in vid_details_df.columns]
        final_df = pd.merge(final_df, vid_details_df,
                            how='left', left_on='videoId', right_on='video.id')

    channel_ids = ','.join(final_df['channelId'].dropna())
    if channel_ids:
        channel_details_df = youtube_channel_details(channel_ids=channel_ids,
                                                     key=key)
        channel_details_df.columns = ['channel.' + x for x in
                                      channel_details_df.columns]

        final_df = pd.merge(final_df, channel_details_df,
                            how='left', left_on='channelId',
                            right_on='channel.id')
    final_df = final_df.drop_duplicates(subset=['videoId'])
    return final_df.reset_index(drop=True)


def set_logging_level(level_or_name):
    """Change the logging level during the session.
    Acceptable values are [0, 10, 20, 30, 40, 50,
    'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR',
    'CRITICAL']
    """
    lvl_names_values = [0, 10, 20, 30, 40, 50,
                        'NOTSET', 'DEBUG', 'INFO',
                        'WARNING', 'ERROR', 'CRITICAL']
    if level_or_name not in lvl_names_values:
        raise ValueError('Please make sure you supply'
                         ' a value from: {}'.format(lvl_names_values))
    logging.getLogger().setLevel(level_or_name)


logging.getLogger().setLevel('INFO')



"""
convert a robots.txt file (or list of file URLs) to a table in a CSV format
============================================================================


    usage: advertools robots [-h] [url ...]
    
    convert a robots.txt file (or list of file URLs) to a table in a CSV format
    
    you can provide a web URL, or a local file URL on your local machine
    e.g. file:///Users/path/to/robots.txt
    
    examples:
    ---------
    
    advertools robots https://www.google.com/robots.txt
    
    multiple robots files:
    
    advertools robots https://www.google.com/robots.txt https://www.google.jo/robots.txt  https://www.google.es/robots.txt
    
    use output redirection ">" to save to a CSV file:
    
    advertools robots https://www.google.com/robots.txt > google_robots.csv
    
    run the function for a long list of robots files saved in a text file (robotslist.txt):
    
    advertools robots < robotslist.txt > multi_robots.csv
    
    positional arguments:
      url         a robots.txt URL (or a list of URLs) (default: None)
    
    optional arguments:
      -h, --help  show this help message and exit
    

download, parse, and save an XML sitemap to a table in a CSV file
==================================================================


    usage: advertools sitemaps [-h] [-r {0,1}] [-s SEPARATOR] [sitemap_url]
    
    download, parse, and save an XML sitemap to a table in a CSV file
    
    positional arguments:
      sitemap_url           the URL of the XML sitemap (regular or sitemap index) (default: None)
    
    optional arguments:
      -h, --help            show this help message and exit
      -r {0,1}, --recursive {0,1}
                            whether or not to fetch sub-sitemaps if it is a sitemap index file (default: 1)
      -s SEPARATOR, --separator SEPARATOR
                            the separator with which to separate columns of the output (default: ,)
    

split a list of URLs into their components: scheme, netloc, path, query, etc.
==============================================================================


    usage: advertools urls [-h] [url_list ...]
    
    split a list of URLs into their components: scheme, netloc, path, query, etc.
    
    positional arguments:
      url_list    a list of URLs to parse (default: None)
    
    optional arguments:
      -h, --help  show this help message and exit
    

crawl a list of known URLs using the HEAD method
=================================================


    usage: advertools headers [-h] [-s [CUSTOM_SETTINGS ...]]
                              [url_list ...] output_file
    
    crawl a list of known URLs using the HEAD method
    return status codes and all available response headers
    
    positional arguments:
      url_list              a list of URLs (default: None)
      output_file           filepath - where to save the output (.jl)
    
    optional arguments:
      -h, --help            show this help message and exit
      -s [CUSTOM_SETTINGS ...], --custom-settings [CUSTOM_SETTINGS ...]
                            settings that modify the behavior of the crawler
                            settings should be separated by spaces, and each setting name and value should
                            be separated by an equal sign '=' without spaces between them
                            
                            example:
                            
                            advertools headers https://example.com example.jl --custom-settings LOG_FILE=logs.log CLOSESPIDER_TIMEOUT=20
                             (default: None)
    

parse, compress and convert a log file to a DataFrame in the .parquet format
=============================================================================


    usage: advertools logs [-h] [-f [FIELDS ...]]
                           log_file output_file errors_file log_format
    
    parse, compress and convert a log file to a DataFrame in the .parquet format
    
    positional arguments:
      log_file              filepath - the log file
      output_file           filepath - where to save the output (.parquet)
      errors_file           filepath - where to save the error lines (.txt)
      log_format            the format of the logs, available defaults are:
                                common, combined, common_with_vhost, nginx_error, apache_error
                                supply a special regex instead if you have a different format
    
    optional arguments:
      -h, --help            show this help message and exit
      -f [FIELDS ...], --fields [FIELDS ...]
                            in case you have a special log format, provide a list of the fields names
                            which will become column names in the parsed compressed file (default: None)
    

perform a reverse DNS lookup on a list of IP addresses
=======================================================


    usage: advertools dns [-h] [ip_list ...]
    
    perform a reverse DNS lookup on a list of IP addresses
    
    positional arguments:
      ip_list     a list of IP addresses (default: None)
    
    optional arguments:
      -h, --help  show this help message and exit
    

generate a table of SEM keywords by supplying a list of products and a list of intent words
============================================================================================


    usage: advertools semkw [-h] [-t [{exact,phrase,modified,broad} ...]]
                            [-l MAX_LEN] [-c {0,1}] [-m {0,1}] [-n CAMPAIGN_NAME]
                            products words
    
    generate a table of SEM keywords by supplying a list of products and a list of intent words
    
    positional arguments:
      products              a file containing the products that you sell, one per line
      words                 a file containing the intent words/phrases that you want to combine with products, one per line
    
    optional arguments:
      -h, --help            show this help message and exit
      -t [{exact,phrase,modified,broad} ...], --match-types [{exact,phrase,modified,broad} ...]
      -l MAX_LEN, --max-len MAX_LEN
                            the number of words that should be combined with products (default: 3)
      -c {0,1}, --capitalize-adgroups {0,1}
                            whether or not to capitalize ad group names in the output file (default: 1)
      -m {0,1}, --order-matters {0,1}
                            do you want combinations and permutations, or just combinations?
                            "buy product" and "product buy" or just "buy product"? (default: 1)
      -n CAMPAIGN_NAME, --campaign-name CAMPAIGN_NAME
    

get stopwords of the selected language
=======================================


    usage: advertools stopwords [-h]
                                {arabic,azerbaijani,bengali,catalan,chinese,croatian,danish,dutch,english,finnish,french,german,greek,hebrew,hindi,hungarian,indonesian,irish,italian,japanese,kazakh,nepali,norwegian,persian,polish,portuguese,romanian,russian,sinhala,spanish,swedish,tagalog,tamil,tatar,telugu,thai,turkish,ukrainian,urdu,vietnamese}
    
    get stopwords of the selected language
    
    positional arguments:
      {arabic,azerbaijani,bengali,catalan,chinese,croatian,danish,dutch,english,finnish,french,german,greek,hebrew,hindi,hungarian,indonesian,irish,italian,japanese,kazakh,nepali,norwegian,persian,polish,portuguese,romanian,russian,sinhala,spanish,swedish,tagalog,tamil,tatar,telugu,thai,turkish,ukrainian,urdu,vietnamese}
    
    optional arguments:
      -h, --help            show this help message and exit
    

get word counts of a text list optionally weighted by a number list
====================================================================


    usage: advertools wordfreq [-h] [-n NUMBER_LIST] [-r REGEX] [-l PHRASE_LEN]
                               [-s [STOPWORDS ...]]
                               [text_list ...]
    
    get word counts of a text list optionally weighted by a number list
    
    words (tokens) can be tokenized using any pattern with the --regex option
    word/phrase lengths can also be modified using the --phrase-len option
    
    positional arguments:
      text_list             a text list, one document (sentence, tweet, etc.) per line (default: None)
    
    optional arguments:
      -h, --help            show this help message and exit
      -n NUMBER_LIST, --number-list NUMBER_LIST
                            filepath - a file containing the number list, one number per line (default: None)
      -r REGEX, --regex REGEX
                            a regex to tokenize words (default: None)
      -l PHRASE_LEN, --phrase-len PHRASE_LEN
                            the phrase (token) length to split words (the `n` in n-grams) (default: 1)
      -s [STOPWORDS ...], --stopwords [STOPWORDS ...]
                            a list of stopwords to exclude when counting, defaults to English stopwords
                            run `advertools stopwords english` to get the stopwords
                            change the language to get other stopwords (default: None)


search for emoji using a regex
===============================


    usage: advertools emoji [-h] regex

    search for emoji using a regex

    positional arguments:
      regex       pattern to search for emoji

    optional arguments:
      -h, --help  show this help message and exit


extract structured entities from a text list; emoji, hashtags, mentions
========================================================================


    usage: advertools extract [-h] {emoji,hashtags,mentions} text_list
    
    extract structured entities from a text list; emoji, hashtags, mentions
    
    positional arguments:
      {emoji,hashtags,mentions}
                            which entity you want to extract
      text_list             filepath - a file containing the text list, one phrase per line
    
    optional arguments:
      -h, --help            show this help message and exit
    

tokenize documents (phrases, keywords, tweets, etc) into token of the desired length
=====================================================================================


    usage: advertools tokenize [-h] [-l LENGTH] [-s SEPARATOR] [text_list ...]
    
    tokenize documents (phrases, keywords, tweets, etc) into token of the desired length
    
    positional arguments:
      text_list             filepath - a file containing the text list, one document (sentence, tweet, etc.) per line (default: None)
    
    optional arguments:
      -h, --help            show this help message and exit
      -l LENGTH, --length LENGTH
                            the length of tokens (the n in n-grams) (default: 1)
      -s SEPARATOR, --separator SEPARATOR
                            the character with which to separate the tokens (default: ,)
    

SEO crawler
============


    usage: advertools crawl [-h] [-l FOLLOW_LINKS] [-d [ALLOWED_DOMAINS ...]]
                            [--exclude-url-params [EXCLUDE_URL_PARAMS ...]]
                            [--include-url-params [INCLUDE_URL_PARAMS ...]]
                            [--exclude-url-regex EXCLUDE_URL_REGEX]
                            [--include-url-regex INCLUDE_URL_REGEX]
                            [--css-selectors [CSS_SELECTORS ...]]
                            [--xpath-selectors [XPATH_SELECTORS ...]]
                            [--custom-settings [CUSTOM_SETTINGS ...]]
                            [url_list ...] output_file
    
    SEO crawler
    
    positional arguments:
      url_list              one or more URLs to crawl (default: None)
      output_file           filepath - where to save the output (.jl)
    
    optional arguments:
      -h, --help            show this help message and exit
      -l FOLLOW_LINKS, --follow-links FOLLOW_LINKS
                            whether or not to follow links encountered on crawled pages (default: 0)
      -d [ALLOWED_DOMAINS ...], --allowed-domains [ALLOWED_DOMAINS ...]
                            while following links, only links on these domains will be followed (default: None)
      --exclude-url-params [EXCLUDE_URL_PARAMS ...]
                            a list of URL parameters to exclude while following links
                            if a link contains any of those parameters, don't follow it
                            setting it to True will exclude links containing any parameter (default: None)
      --include-url-params [INCLUDE_URL_PARAMS ...]
                            a list of URL parameters to include while following links
                            if a link contains any of those parameters, follow it
                            having the same parmeters to include and exclude raises an error (default: None)
      --exclude-url-regex EXCLUDE_URL_REGEX
                            a regular expression of a URL pattern to exclude while following links
                            if a link matches the regex don't follow it (default: None)
      --include-url-regex INCLUDE_URL_REGEX
                            a regular expression of a URL pattern to include while following links
                            if a link matches the regex follow it (default: None)
      --css-selectors [CSS_SELECTORS ...]
                            a dictionary mapping names to CSS selectors
                            the names will become column headers, and the selectors will be used to extract the required data/content (default: None)
      --xpath-selectors [XPATH_SELECTORS ...]
                            a dictionary mapping names to XPath selectors.
                            the names will become column headers, and the selectors will be used to extract the required data/content (default: None)
      --custom-settings [CUSTOM_SETTINGS ...]
                            a dictionary of optional custom settings that you might want to
                            add to the spider's functionality.
                            there are over 170 settings for all kinds of options
                            for details please refer to the spider settings:
                            https://docs.scrapy.org/en/latest/topics/settings.html (default: None)
    
    Examples:
    ---------
    
    crawl a website starting from its home page:
    
    advertools crawl https://examle.com example_output.jl --follow-links 1
    
    crawl a list of pages (list mode):
    
    advertools crawl url_1 url_2 url_3 example_output.jl
    
    OR if you have a long list in a file (url_list.txt):
    
    advertools crawl < url_list.txt example_output.jl
    
    stop crawling after having processed 1,000 pages:
    
    advertools crawl https://examle.com example_output.jl --follow-links 1 --custom-settings CLOSESPIDER_PAGECOUNT=1000

"""
import argparse
import platform
import socket
import sys
from concurrent import futures
from textwrap import dedent

import pandas as pd

import advertools as adv
from advertools import __version__

system = platform.system()

_max_workers = 60 if system == 'Darwin' else 600

_entity_dict = {
    'emoji': adv.emoji.EMOJI,
    'hashtags': adv.regex.HASHTAG,
    'mentions': adv.regex.MENTION,
}


def _make_headline(text, indent=0):
    len_ = len(text)
    top = '═' * len_
    bottom = '═' * len_
    return '\n'.join([(' ' * indent) + top,
                      (' ' * indent) + text,
                      (' ' * indent) + bottom])


epilog = _make_headline('full documentation at https://bit.ly/adv_docs')


def _split_options(options):
    opsplit = [op.split('=', maxsplit=1) for op in options]
    d = {k: v for k, v in opsplit}
    for k, v in d.items():
        if v == 'True':
            d[k] = True
        if v == 'False':
            d[k] = False
    return d


def _format_df(df, head=10, precision=1):
    for col in df:
        if df[col].dtype == int:
            df[col] = [format(n, ',') for n in df[col]]
        if df[col].dtype == float:
            df[col] = [format(n, f',.{precision}f') for n in df[col]]
    return df.head(head)


def _single_request(ip):
    try:
        hostname, aliaslist, ipaddrlist = socket.gethostbyaddr(ip)
        return [ip, hostname, '@@'.join(aliaslist), '@@'.join(ipaddrlist)]
    except Exception as e:
        return [ip, None, None, None, str(e)]



def _cli_reverse_dns_lookup(ip_list, max_workers=600):
    socket.setdefaulttimeout(8)
    count_df = (pd.Series(ip_list)
                .value_counts()
                .reset_index())
    count_df.columns = ['ip_address', 'count']
    count_df['cum_count'] = count_df['count'].cumsum()
    count_df['perc'] = count_df['count'].div(count_df['count'].sum())
    count_df['cum_perc'] = count_df['perc'].cumsum()

    hosts = []
    with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        for host in executor.map(_single_request, count_df['ip_address']):
            hosts.append(host)

    df = pd.DataFrame(hosts)
    columns = ['ip', 'hostname', 'aliaslist', 'ipaddrlist', 'errors']
    if df.shape[1] == 4:
        columns = columns[:-1]
    df.columns = columns

    final_df = pd.merge(count_df, df, left_on='ip_address',
                        right_on='ip', how='left').drop('ip', axis=1)
    return final_df


class RawTextDefArgFormatter(argparse.RawTextHelpFormatter,
                             argparse.ArgumentDefaultsHelpFormatter):
    pass


def main():
    parser = argparse.ArgumentParser(
        prog='advertools',
        formatter_class=RawTextDefArgFormatter,
        epilog=epilog)
    parser.add_argument('-v', '--version', action='version',
                        version=f'advertools {__version__}')

    subparsers = parser.add_subparsers(
        help='for help select a command and run: `advertools <command> --help`')

    # robots --------------------------

    def robots(args):
        if sys.stdin.isatty():
            url = args.url
        else:
            url = [u.strip() for u in sys.stdin.read().split()]
        if not url:
            print("error: please provide a value for url", file=sys.stderr)
            sys.exit(1)
        print(adv.robotstxt_to_df(url).to_csv(index=False))

    robots_parser = subparsers.add_parser(
        'robots',
        formatter_class=RawTextDefArgFormatter,
        epilog=epilog,
        description='''convert a robots.txt file (or list of file URLs) to a table in a CSV format

you can provide a web URL, or a local file URL on your local machine
e.g. file:///Users/path/to/robots.txt

examples:
---------

advertools robots https://www.google.com/robots.txt

multiple robots files:

advertools robots https://www.google.com/robots.txt https://www.google.jo/robots.txt  https://www.google.es/robots.txt

use output redirection ">" to save to a CSV file:

advertools robots https://www.google.com/robots.txt > google_robots.csv

run the function for a long list of robots files saved in a text file (robotslist.txt):

advertools robots < robotslist.txt > multi_robots.csv
''')
    robots_parser.add_argument(
        'url', nargs='*',
        help='a robots.txt URL (or a list of URLs)')
    robots_parser.set_defaults(func=robots)

    # sitemaps --------------------------

    def sitemaps(args):
        if sys.stdin.isatty():
            sitemap_url = args.sitemap_url
        else:
            sitemap_url = sys.stdin.read().strip()
        if not sitemap_url:
            print("error: please provide a value for sitemap_url",
                  file=sys.stderr)
            sys.exit(1)
        sitemap_df = adv.sitemap_to_df(sitemap_url,
                                       recursive=bool(args.recursive))
        for col in sitemap_df:
            if col in ['news', 'image', 'news_publication']:
                del sitemap_df[col]
        print(sitemap_df.to_csv(sep=args.separator, index=False))

    sitemaps_parser = subparsers.add_parser(
        'sitemaps', formatter_class=RawTextDefArgFormatter, epilog=epilog,
        description='download, parse, and save an XML sitemap to a table in a CSV file')
    sitemaps_parser.add_argument(
        'sitemap_url', nargs='?',
        help='the URL of the XML sitemap (regular or sitemap index)')
    sitemaps_parser.add_argument(
        '-r', '--recursive', type=int, choices=[0, 1],
        help='whether or not to fetch sub-sitemaps if it is a sitemap index file',
        default=1, required=False)
    sitemaps_parser.add_argument(
        '-s', '--separator', type=str, default=',',
        help='the separator with which to separate columns of the output'
    )

    sitemaps_parser.set_defaults(func=sitemaps)

    # urls --------------------------

    def urls(args):
        if sys.stdin.isatty():
            url_list = args.url_list
        else:
            url_list = [url.strip() for url in sys.stdin.read().split()]
        if not url_list:
            print("error: please provide a value for url_list",
                  file=sys.stderr)
            sys.exit(1)
        print(adv.url_to_df(url_list).to_csv(index=False))

    urls_parser = subparsers.add_parser(
        'urls', formatter_class=RawTextDefArgFormatter, epilog=epilog,
        description='split a list of URLs into their components: scheme, netloc, path, query, etc.')
    urls_parser.add_argument(
        'url_list',  nargs='*',
        help='a list of URLs to parse')
    urls_parser.set_defaults(func=urls)

    # headers --------------------------

    def headers(args):
        if sys.stdin.isatty():
            url_list = args.url_list
        else:
            url_list = [url.strip() for url in sys.stdin.read().split()]
        if not url_list:
            print("error: please provide a value for url_list",
                  file=sys.stderr)
            sys.exit(1)
        if args.custom_settings:
            args.custom_settings = _split_options(args.custom_settings)
        adv.crawl_headers(url_list=url_list,
                          output_file=args.output_file,
                          custom_settings=args.custom_settings)

    headers_parser = subparsers.add_parser(
        'headers', formatter_class=RawTextDefArgFormatter, epilog=epilog,
        description='''crawl a list of known URLs using the HEAD method
return status codes and all available response headers''')
    headers_parser.add_argument(
        'url_list', nargs='*',
        help='a list of URLs')
    headers_parser.add_argument(
        'output_file', type=str,
        help='filepath - where to save the output (.jl)')
    headers_parser.add_argument(
        '-s', '--custom-settings', nargs='*',
        help='''settings that modify the behavior of the crawler
settings should be separated by spaces, and each setting name and value should
be separated by an equal sign '=' without spaces between them

example:

advertools headers https://example.com example.jl --custom-settings LOG_FILE=logs.log CLOSESPIDER_TIMEOUT=20
''')
    headers_parser.set_defaults(func=headers)

    # logs --------------------------

    def logs(args):
        adv.logs_to_df(log_file=args.log_file,
                       output_file=args.output_file,
                       errors_file=args.errors_file,
                       log_format=args.log_format,
                       fields=args.fields)
        print(f'saved to {args.output_file}')

    logs_parser = subparsers.add_parser(
        'logs', epilog=epilog, formatter_class=RawTextDefArgFormatter,
        description='parse, compress and convert a log file to a DataFrame in the .parquet format')
    logs_parser.add_argument(
        'log_file', type=str, help='filepath - the log file')
    logs_parser.add_argument(
        'output_file', type=str,
        help='filepath - where to save the output (.parquet)')
    logs_parser.add_argument(
        'errors_file', type=str,
        help='filepath - where to save the error lines (.txt)')
    logs_parser.add_argument(
        'log_format', type=str, default='common',
        help='''the format of the logs, available defaults are:
    common, combined, common_with_vhost, nginx_error, apache_error
    supply a special regex instead if you have a different format''')
    logs_parser.add_argument(
        '-f', '--fields', type=str, nargs='*',
        help='''in case you have a special log format, provide a list of the fields names
which will become column names in the parsed compressed file''')
    logs_parser.set_defaults(func=logs)

    # dns --------------------------

    def dns(args):
        if sys.stdin.isatty():
            ip_list = args.ip_list
        else:
            ip_list = sys.stdin.read().split()
        if not ip_list:
            print('please provide a value for ip_list', file=sys.stderr)
            sys.exit(1)
        host_df = _cli_reverse_dns_lookup(ip_list)
        print(host_df.to_csv(index=False))

    dns_parser = subparsers.add_parser(
        'dns', epilog=epilog, formatter_class=RawTextDefArgFormatter,
        description='perform a reverse DNS lookup on a list of IP addresses')
    dns_parser.add_argument(
        'ip_list', nargs='*',
        help='a list of IP addresses')
    dns_parser.set_defaults(func=dns)

    # keywords --------------------------

    def semkw(args):
        prod_list = [line.strip() for line in args.products.readlines()]
        word_list = [line.strip() for line in args.words.readlines()]
        kw_df = adv.kw_generate(prod_list, word_list,
                                max_len=args.max_len,
                                match_types=args.match_types,
                                capitalize_adgroups=bool(args.capitalize_adgroups),
                                order_matters=bool(args.order_matters),
                                campaign_name=args.campaign_name)
        print(kw_df.to_csv(index=False))

    kwds_parser = subparsers.add_parser(
        'semkw', formatter_class=RawTextDefArgFormatter, epilog=epilog,
        description='generate a table of SEM keywords by supplying a list of products and a list of intent words')
    kwds_parser.add_argument(
        'products', type=open,
        help='a file containing the products that you sell, one per line')
    kwds_parser.add_argument(
        'words', type=open,
        help='a file containing the intent words/phrases that you want to combine with products, one per line')
    kwds_parser.add_argument('-t', '--match-types', type=str, nargs='*',
                             default=['exact', 'phrase'],
                             choices=['exact', 'phrase', 'modified', 'broad'])
    kwds_parser.add_argument(
        '-l', '--max-len', type=int, default=3,
        help='the number of words that should be combined with products')
    kwds_parser.add_argument(
        '-c', '--capitalize-adgroups', type=int, default=1, choices=[0, 1],
        help='whether or not to capitalize ad group names in the output file')
    kwds_parser.add_argument(
        '-m', '--order-matters', type=int, default=1, choices=[0, 1],
        help='''\
do you want combinations and permutations, or just combinations?
"buy product" and "product buy" or just "buy product"?''')
    kwds_parser.add_argument('-n', '--campaign-name', type=str,
                             default='SEM_campaign')
    kwds_parser.set_defaults(func=semkw)

    # stopwords --------------------------

    def stopwords(args):
        print(*sorted(adv.stopwords[args.language]), sep='\n')

    stopwords_parser = subparsers.add_parser(
        'stopwords',
        formatter_class=RawTextDefArgFormatter,
        epilog=epilog,
        description='get stopwords of the selected language')
    stopwords_parser.add_argument(
        'language', type=str, choices=adv.stopwords.keys())
    stopwords_parser.set_defaults(func=stopwords)

    # word_freq --------------------------

    def word_freq(args):
        if sys.stdin.isatty():
            text_list = args.text_list
        else:
            text_list = sys.stdin.read().split()
        if not text_list:
            print('please provide a value for text_list', file=sys.stderr)
            sys.exit(1)

        # text_list = pd.read_csv(args.text_list, header=None)[0]
        if args.number_list is not None:
            num_list = pd.read_csv(args.number_list, header=None)[0]
        else:
            num_list = [1 for _ in range(len(text_list))]
        wordfreq_df = adv.word_frequency(
            text_list=text_list,
            num_list=num_list,
            regex=args.regex,
            phrase_len=args.phrase_len,
            rm_words=args.stopwords or adv.stopwords['english'])
        print(wordfreq_df.to_csv(index=False))

    wordfreq_parser = subparsers.add_parser(
        'wordfreq', epilog=epilog, formatter_class=RawTextDefArgFormatter,
        description='''get word counts of a text list optionally weighted by a number list

words (tokens) can be tokenized using any pattern with the --regex option
word/phrase lengths can also be modified using the --phrase-len option''')
    wordfreq_parser.add_argument(
        'text_list', type=str, nargs='*',
        help='a text list, one document (sentence, tweet, etc.) per line')
    wordfreq_parser.add_argument(
        '-n', '--number-list', type=str, required=False,
        help='filepath - a file containing the number list, one number per line')
    wordfreq_parser.add_argument(
        '-r', '--regex', type=str, required=False,
        help='a regex to tokenize words')
    wordfreq_parser.add_argument(
        '-l', '--phrase-len', type=int, required=False, default=1,
        help='the phrase (token) length to split words (the `n` in n-grams)')
    wordfreq_parser.add_argument(
        '-s', '--stopwords', type=str, nargs='*', required=False,
        help=dedent('''\
    a list of stopwords to exclude when counting, defaults to English stopwords
    run `advertools stopwords english` to get the stopwords
    change the language to get other stopwords'''))
    wordfreq_parser.set_defaults(func=word_freq)

    # emoji --------------------------

    def emoji(args):
        emoji_df = adv.emoji_search(args.regex)
        print(emoji_df.to_csv(index=False))

    emoji_parser = subparsers.add_parser(
        'emoji', formatter_class=RawTextDefArgFormatter, epilog=epilog,
        description='search for emoji using a regex')
    emoji_parser.add_argument('regex', type=str,
                              help='pattern to search for emoji')
    emoji_parser.set_defaults(func=emoji)

    # extract --------------------------

    def extract(args):
        text_list = pd.read_csv(args.text_list, header=None)[0]
        extracted = adv.extract(
            text_list=text_list, key_name=args.entity.rstrip('s'),
            regex=_entity_dict[args.entity])
        extracted_subset = {key: val for key, val in extracted.items()
                            if len(val) == len(text_list)}
        for key, val in extracted_subset.items():
            if isinstance(val[0], list):
                extracted_subset[key] = [', '.join(v) for v in val]
        print(file=sys.stderr)
        print(pd.DataFrame(extracted['overview'], index=['>']), file=sys.stderr)
        print(file=sys.stderr)
        print(f'top 15 {args.entity}:\n', file=sys.stderr)
        top_df = pd.DataFrame(extracted[[k for k in extracted
                                         if k.startswith('top_')][0]],
                              columns=[args.entity, 'count'])
        print(top_df[:15], file=sys.stderr)
        freq_df = pd.DataFrame(extracted[[k for k in extracted
                                          if k.endswith('_freq')][0]])
        freq_df.columns = ['number of ' + args.entity, 'count']
        print(f'\n{args.entity} frequency:\n', file=sys.stderr)
        print(freq_df, file=sys.stderr)
        print(pd.DataFrame(extracted_subset).to_csv(index=False))

    extract_parser = subparsers.add_parser(
        'extract', formatter_class=RawTextDefArgFormatter, epilog=epilog,
        description=f'extract structured entities from a text list; {", ".join(_entity_dict.keys())}')
    extract_parser.add_argument(
        'entity', help='which entity you want to extract',
        choices=_entity_dict.keys())
    extract_parser.add_argument(
        'text_list',
        help='filepath - a file containing the text list, one phrase per line')

    extract_parser.set_defaults(func=extract)

    # tokenize --------------------------

    def tokenize(args):
        if sys.stdin.isatty():
            text_list = args.text_list
        else:
            text_list = [text.strip() for text in sys.stdin.readlines()]
        if not text_list:
            print("error: please provide a value for text_list", file=sys.stderr)
            sys.exit(1)
        tokenized = adv.word_tokenize(text_list, args.length)
        print(*[args.separator.join(t) for t in tokenized], sep='\n')

    token_parser = subparsers.add_parser(
        'tokenize', formatter_class=RawTextDefArgFormatter, epilog=epilog,
        description='tokenize documents (phrases, keywords, tweets, etc) into token of the desired length')
    token_parser.add_argument(
        'text_list', type=str, nargs='*',
        help='filepath - a file containing the text list, one document (sentence, tweet, etc.) per line')
    token_parser.add_argument(
        '-l', '--length', type=int, default=1,
        help='the length of tokens (the n in n-grams)')
    token_parser.add_argument(
        '-s', '--separator', type=str, default=",",
        help='the character with which to separate the tokens')

    token_parser.set_defaults(func=tokenize)

    def crawl(args):
        if sys.stdin.isatty():
            url_list = args.url_list
        else:
            url_list = [url.strip() for url in sys.stdin.read().split()]
        if not url_list:
            print("error: please provide a value for url_list", file=sys.stderr)
            return

        if args.exclude_url_params:
            if args.exclude_url_params in ['True', 1, '1', 'true']:
                exclude_url_params = True
            else:
                exclude_url_params = args.exclude_url_params
        else:
            exclude_url_params = None
        if args.xpath_selectors:
            args.xpath_selectors = _split_options(args.xpath_selectors)
        if args.css_selectors:
            args.css_selectors = _split_options(args.css_selectors)
        if args.custom_settings:
            args.custom_settings = _split_options(args.custom_settings)
        adv.crawl(url_list=url_list,
                  output_file=args.output_file,
                  allowed_domains=args.allowed_domains,
                  css_selectors=args.css_selectors,
                  xpath_selectors=args.xpath_selectors,
                  follow_links=bool(int(args.follow_links)),
                  custom_settings=args.custom_settings,
                  exclude_url_params=exclude_url_params,
                  include_url_params=args.include_url_params,
                  exclude_url_regex=args.exclude_url_regex,
                  include_url_regex=args.include_url_regex)

    crawl_parser = subparsers.add_parser(
        'crawl', formatter_class=RawTextDefArgFormatter,
        epilog="""
Examples:
---------

crawl a website starting from its home page:

advertools crawl https://examle.com example_output.jl --follow-links 1

crawl a list of pages (list mode):

advertools crawl url_1 url_2 url_3 example_output.jl

OR if you have a long list in a file (url_list.txt):

advertools crawl < url_list.txt example_output.jl

stop crawling after having processed 1,000 pages:

advertools crawl https://examle.com example_output.jl --follow-links 1 --custom-settings CLOSESPIDER_PAGECOUNT=1000
""" + epilog,
        description='SEO crawler')
    crawl_parser.add_argument(
        'url_list', nargs='*', help='one or more URLs to crawl')
    crawl_parser.add_argument(
        'output_file',
        help='filepath - where to save the output (.jl)')
    crawl_parser.add_argument(
        '-l', '--follow-links', default=0, type=int,
        help='whether or not to follow links encountered on crawled pages')
    crawl_parser.add_argument(
        '-d', '--allowed-domains', type=str, nargs='*',
        help='while following links, only links on these domains will be followed')
    crawl_parser.add_argument(
        '--exclude-url-params', type=str, nargs='*',
        help='''a list of URL parameters to exclude while following links
if a link contains any of those parameters, don't follow it
setting it to True will exclude links containing any parameter''')
    crawl_parser.add_argument(
        '--include-url-params', type=str, nargs='*',
        help='''a list of URL parameters to include while following links
if a link contains any of those parameters, follow it
having the same parmeters to include and exclude raises an error''')
    crawl_parser.add_argument(
        '--exclude-url-regex',  type=str,
        help='''a regular expression of a URL pattern to exclude while following links
if a link matches the regex don't follow it''')
    crawl_parser.add_argument(
        '--include-url-regex', type=str,
        help='''a regular expression of a URL pattern to include while following links
if a link matches the regex follow it''')
    crawl_parser.add_argument(
        '--css-selectors', nargs='*', type=str,
        help='''a dictionary mapping names to CSS selectors
the names will become column headers, and the selectors will be used to extract the required data/content''')
    crawl_parser.add_argument(
        '--xpath-selectors', nargs='*', type=str,
        help='''a dictionary mapping names to XPath selectors.
the names will become column headers, and the selectors will be used to extract the required data/content''')
    crawl_parser.add_argument(
        '--custom-settings', type=str, nargs='*',
        help='''a dictionary of optional custom settings that you might want to
add to the spider's functionality.
there are over 170 settings for all kinds of options
for details please refer to the spider settings:
https://docs.scrapy.org/en/latest/topics/settings.html''')

    crawl_parser.set_defaults(func=crawl)
    return parser


parser = main()
args = parser.parse_args()
args.func(args)



#!/usr/bin/env python

"""The setup script."""

import os

from setuptools import find_packages, setup

os.listdir()
with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "pandas>=1.1.0",
    "pyasn1>=0.4",
    "scrapy>=2.5.0",
    "twython>=3.8.0",
    "pyarrow>=5.0.0",
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

setup(
    author="Elias Dabbas",
    author_email="eliasdabbas@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="Productivity and analysis tools for online marketing",
    entry_points={
        "console_scripts": [
            "advertools=advertools.cli:main",
            "adv=advertools.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="advertising marketing search-engine-optimization adwords "
    "seo sem bingads keyword-research",
    name="advertools",
    packages=find_packages(include=["advertools"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/eliasdabbas/advertools",
    version="0.14.1",
    zip_safe=False,
)



import logging
import os
from itertools import product

import pandas as pd
import pytest
from advertools.serp import (SERP_GOOG_VALID_VALS, SERP_YTUBE_VALID_VALS,
                             YOUTUBE_TOPIC_IDS, YOUTUBE_VID_CATEGORY_IDS,
                             _dict_product, serp_goog, serp_youtube,
                             set_logging_level, youtube_channel_details,
                             youtube_video_details)

goog_cse_cx = os.environ.get('GOOG_CSE_CX')
goog_cse_key = os.environ.get('GOOG_CSE_KEY')
youtube_key = os.environ.get('GOOG_CSE_KEY')

skip_api_tests = pytest.mark.skipif(os.environ.get('ADV_TEST_OFFLINE'),
                                    reason='Run all except API dependents')


def test_dict_product_produces_correct_result():
    d = {'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [10, 20]}
    dp_exp = [
        {'a': 1, 'b': 4, 'c': 10},
        {'a': 1, 'b': 4, 'c': 20},
        {'a': 1, 'b': 5, 'c': 10},
        {'a': 1, 'b': 5, 'c': 20},
        {'a': 1, 'b': 6, 'c': 10},
        {'a': 1, 'b': 6, 'c': 20},
        {'a': 2, 'b': 4, 'c': 10},
        {'a': 2, 'b': 4, 'c': 20},
        {'a': 2, 'b': 5, 'c': 10},
        {'a': 2, 'b': 5, 'c': 20},
        {'a': 2, 'b': 6, 'c': 10},
        {'a': 2, 'b': 6, 'c': 20},
        {'a': 3, 'b': 4, 'c': 10},
        {'a': 3, 'b': 4, 'c': 20},
        {'a': 3, 'b': 5, 'c': 10},
        {'a': 3, 'b': 5, 'c': 20},
        {'a': 3, 'b': 6, 'c': 10},
        {'a': 3, 'b': 6, 'c': 20}
    ]
    dp_res = _dict_product(d)
    for d_res in dp_res:
        assert d_res in dp_exp


def test_dict_product_return_correct_types():
    d = {'a': [1], 'b': [10, 20, 30], 'c': (4, 5, 6)}
    dp = _dict_product(d)
    assert isinstance(dp, list)
    assert [isinstance(x, dict) for x in dp]
    assert len(dp) == len(list(product(*d.values())))


# Google search tests:
def test_serp_goog_raises_error_on_invalid_args():
    with pytest.raises(ValueError):
        for val in SERP_GOOG_VALID_VALS:
            params = {val: 'WRONG VALUE'}
            serp_goog(q='q', cx='cx', key='key', **params)


@skip_api_tests
def test_serp_goog_return_correct_result():
    result = serp_goog(q='testing hotels', cx=goog_cse_cx,
                       key=goog_cse_key, searchType=['image', None])
    assert isinstance(result, pd.core.frame.DataFrame)
    assert 'title' in result
    assert 'image' in result
    assert len(result) == 20


@skip_api_tests
def test_serp_goog_handles_no_search_results():
    q = 'aquerythatdoesntgetrezultssss'
    result = serp_goog(q=q, cx=goog_cse_cx, key=goog_cse_key,
                       cr='countryRU', hl='zh-TW', gl='nf')
    assert len(result) == 1
    assert result['searchTerms'].values[0] == q


@skip_api_tests
def test_serp_raises_error_on_wrong_key():
    with pytest.raises(Exception):
        serp_goog(q='test credit cart', cx=goog_cse_cx, key='wrong key')


@skip_api_tests
def test_serp_goog_restult_df_contains_all_provided_params():
    keys_vals_to_test = {k: list(SERP_GOOG_VALID_VALS[k])[0] for k in SERP_GOOG_VALID_VALS}
    for key, val in keys_vals_to_test.items():
        result = serp_goog(cx=goog_cse_cx, key=goog_cse_key, q='fashion', **{key: val})
        if key == 'searchType':
            continue
        assert key in result.columns
        if key == 'filter':
            val = str(val)
        assert result[key].iloc[0] == val


# YouTube search tests:
def test_serp_youtube_raises_error_on_invalid_args():
    with pytest.raises(ValueError):
        for val in SERP_YTUBE_VALID_VALS:
            params = {val: 'WRONG VALUE'}
            serp_youtube(q='q', key='key', **params)


@skip_api_tests
def test_serp_youtube_return_correct_result():
    result = serp_youtube(q=['testing hotels', 'testing computers'],
                          key=youtube_key, order='date')
    assert isinstance(result, pd.core.frame.DataFrame)
    assert 'title' in result
    assert 'rank' in result
    assert len(result) <= 10


@skip_api_tests
def test_serp_youtube_handles_no_search_results():
    q = 'aquerythatdoesntgetrezultssss'
    result = serp_youtube(q=q, key=youtube_key,
                          relevanceLanguage='ar',
                          channelId='UCN1Q_Z0QgTKMimCv6G0IkKQ')
    assert len(result) == 1
    assert result['q'].values[0] == q
    assert result['rank'].isna().values[0]


@skip_api_tests
def test_serp_youtube_raises_type_video_error():
    with pytest.raises(Exception):
        serp_youtube(key=youtube_key, videoEmbeddable=True)


@skip_api_tests
def test_serp_youtube_raises_response_error():
    with pytest.raises(Exception):
        serp_youtube(key=youtube_key, publishedAfter='wrong date fmt')


def test_correctly_changing_log_levels():
    lvl_names_values = [0, 10, 20, 30, 40, 50]
    for level in lvl_names_values:
        set_logging_level(level)
        assert logging.getLogger().level == level
    with pytest.raises(ValueError):
        set_logging_level('WRONG VALUE')


@skip_api_tests
def test_youtube_video_details_raises_error():
    with pytest.raises(Exception):
        youtube_video_details(key='WRONG KEY',
                              vid_ids='wrong ID')


@skip_api_tests
def test_youtube_channel_details_raises_error():
    with pytest.raises(Exception):
        youtube_channel_details(key='WRONG KEY',
                                channel_ids='wrong ID')



"""
.. _regex:

Regular Expressions for Extracting Structured Entities
======================================================

A collection of regular expressions for use in different contexts.
Each one is available in two formats:

* REGEX_RAW: (HASHTAG_RAW, MENTION_RAW, etc.) raw string only, for sharing
             and combining with other regexes
* REGEX: (HASHTAG, MENTION, etc.) compiled regex, readable, and annotated

Based on Unicode database v11.0.0

URL regex from Regular Expressions Cookbook 2nd Ed. O'Reilly
"""

__all__ = ['APOSTROPHE', 'BRACKET', 'COLON', 'COMMA', 'CURRENCY',
           'CURRENCY_RAW', 'EXCLAMATION', 'EXCLAMATION_MARK', 'FULL_STOP',
           'HASHTAG', 'HASHTAG_RAW', 'MENTION', 'MENTION_RAW', 'PAREN',
           'QUESTION', 'QUESTION_MARK', 'QUESTION_MARK_NEG_RAW',
           'QUESTION_MARK_RAW', 'QUESTION_RAW', 'QUOTE',
           'SENTENCE_END', 'WORD_DELIM', 'URL', 'URL_RAW']

import re


# word delimiters used to extract words
QUOTE = r'["«»‘’‚‛“”„‟‹›❮❯⹂〝〞〟＂]'

EXCLAMATION = r'[!¡՜߹᥄‼⁈⁉︕﹗！𖺚𞥞]'

FULL_STOP = r'[.։۔܁܂።᙮᠃᠉⳹⳾⸼。꓿꘎꛳︒﹒．｡𖫵𖺘𛲟𝪈]'

COMMA = r'[,՝،߸፣᠂᠈⸲⸴⹁⹉⹌、꓾꘍꛵︐︑﹐﹑，､𑑍𖺗𝪇]'

BRACKET = (r'[[]{}⁅⁆〈〉❬❭❰❱❲❳❴❵⟦⟧⟨⟩⟪⟫⟬⟭⦃⦄⦇⦈⦉⦊⦋⦌⦍⦎⦏⦐⦑⦒⦓⦔⦕⦖⦗⦘⧼⧽⸂⸃⸄⸅⸉⸊⸌⸍⸜⸝⸢⸣⸤⸥⸦⸧〈〉'
           r'《》「」『』【】〔〕〖〗〘〙〚〛︗︷︸︹︺︻︼︽︾︿﹀﹁﹂﹃﹄﹇﹈'
           r'﹛﹜﹝﹞［］｛｝｢｣]')

COLON = r'[:;؛܃܄܅܆܇܈܉፤፥፦᠄⁏⁝⸵꛴꛶︓︔﹔﹕：；𒑱𒑲𒑳𒑴𝪉𝪊]'

PAREN = r'[()⁽⁾₍₎❨❩❪❫⟮⟯⦅⦆⸨⸩﴾﴿︵︶﹙﹚（）｟｠𝪋]'

APOSTROPHE = r'["\'ʼˮ՚ߴߵ＇"]'

EXCLAMATION_MARK_RAW = r'[!¡՜߹᥄‼⁈⁉︕﹗！𖺚𞥞]'

EXCLAMATION_MARK = re.compile(
    r"""[!¡՜߹᥄‼⁈⁉︕﹗！𖺚𞥞]
        # Unicode characters named exclamation mark
    """, re.VERBOSE)

EXCLAMATION_MARK_NEG_RAW = r'[^!¡՜߹᥄‼⁈⁉︕﹗！𖺚𞥞]'

QUESTION_MARK_RAW = r'[?¿;՞؟፧᥅⁇⁈⁉⳺⳻⸮꘏꛷︖﹖？𑅃𞥟' + r'ʔ‽' + r']'

QUESTION_MARK = re.compile(
    r"""[?¿;՞؟፧᥅⁇⁈⁉⳺⳻⸮꘏꛷︖﹖？𑅃𞥟ʔ‽]
         # Unicode characters named question mark
    """, re.VERBOSE)

QUESTION_MARK_NEG_RAW = r'[^?;՞؟፧᥅⁇⁈⁉⳺⳻⸮꘏꛷︖﹖？𑅃𞥟ʔ‽]'

WORD_DELIM = r'[' + r''.join([x.strip('[]')
                              for x in [QUOTE, EXCLAMATION, QUESTION_MARK_RAW,
                                        FULL_STOP, COMMA, BRACKET, COLON,
                                        APOSTROPHE + PAREN]]) + r']'

SENTENCE_END = r'[' + r''.join([x.strip('[]')
                                for x in [EXCLAMATION, FULL_STOP,
                                          QUESTION_MARK_RAW]]) + r']'

HASHTAG_RAW = r'(?i)(?<!\w)([＃#]\w+)'

HASHTAG = re.compile(
    r"""
    (?i)        # case-insensitive mode
    (?<!\w)     # hashtag not preceded by word character
    ([＃#]\w+)  # one of two hashtag characters
    """, re.VERBOSE)

MENTION_RAW = r'(?ix)(?<!\w)([@＠][a-z0-9_]+)\b'

MENTION = re.compile(
    r"""(?i)     # case-insensitive
    (?<!\w)      # word character doesn't precede mention
    ([@＠]       # either of two @ signs
    [a-z0-9_]+)  # A to Z, numbers and underscores only
    \b           # end with a word boundary
    """, re.VERBOSE)

CURRENCY_RAW = r'[$¢£¤¥֏؋৲৳৻૱௹฿៛₠₡₢₣₤₥₦₧₨₩₪₫€₭₮₯₰₱₲₳₴₵₶₷₸₹₺₻₼₽₾₿﷼﹩＄￠￡￥￦]'

CURRENCY = re.compile(
    r"""[$¢£¤¥֏؋৲৳৻૱௹฿៛₠₡₢₣₤₥₦₧₨₩₪₫€₭₮₯₰₱₲₳₴₵₶₷₸₹₺₻₼₽₾₿﷼﹩＄￠￡￥￦]
         # Unicode characters under the category is "Sc"
    """, re.VERBOSE)


EXCLAMATION_RAW = (r'(?i)(?:(?:(?<={})(?:{}*)\s+|^)|(?=¡))(¡?{}+?{}+)'
                   .format(SENTENCE_END, QUOTE,
                           EXCLAMATION_MARK_NEG_RAW,
                           EXCLAMATION_MARK_RAW.replace('¡', '')))


EXCLAMATION = re.compile(r"""
    (?i)           # case insensitive
    (?:
    (?:(?<={s})    # beginning of string, or assert current position
                   #   preceded by a SENTENCE_END character
    (?:{q}*)       # optional quote character(s)
    \s+|^)         # one or more spaces
    |(?=¡))        # or assert current position is "¡" 
    (¡?{neg}+?     # optional Spanish exclamation mark, then one or more
                   #   non-SENTENCE_END characters
     {raw}+)       # one or more exclamation mark characters excluding "¡"
    """.format(s=SENTENCE_END,
               q=QUOTE,
               neg=SENTENCE_END.replace('[', '[^'),
               raw=EXCLAMATION_MARK_RAW.replace('¡', '')),
    re.VERBOSE)


QUESTION_RAW = (r'(?i)(?:(?:(?<={})(?:{}*)\s+|^)|(?=¿))(¿?{}+?{}+)'
                .format(SENTENCE_END, QUOTE,
                        QUESTION_MARK_NEG_RAW,
                        QUESTION_MARK_RAW.replace('¿', '')))

QUESTION = re.compile(r"""
    (?i)           # case insensitive
    (?:
    (?:(?<={s})    # beginning of string, or assert current position
                   #   preceded by a SENTENCE_END character
    (?:{q}*)       # optional quote character(s)
    \s+|^)         # one or more spaces
    |(?=¿))        # or assert current position is "¿" 
    (¿?{neg}+?     # optional Spanish question mark, then one or more
                   #   non-SENTENCE_END characters
     {raw}+)       # one or more question mark characters excluding "¿"
    """.format(s=SENTENCE_END,
               q=QUOTE,
               neg=SENTENCE_END.replace('[', '[^'),
               raw=QUESTION_MARK_RAW.replace('¿', '')),
    re.VERBOSE)

URL_RAW = (r'(?xi)\b(?:(?:https?|ftp|file)://|www\.|ftp\.)'
           r'(?:\([-A-Z0-9+&@#/%=~_|$?!:,.]*\)|'
           r'[-A-Z0-9+&@#/%=~_|$?!:,.])*'
           r'(?:\([-A-Z0-9+&@#/%=~_|$?!:,.]*\)|'
           r'[A-Z0-9+&@#/%=~_|$])')

URL = re.compile(r"""
    (?xi)                                     # case-insensitive / verbose
    \b(?:(?:https?|ftp|file)://|www\.|ftp\.)  # starts w/ http(s), ftp or www
      (?:\([-A-Z0-9+&@#/%=~_|$?!:,.]*\)|      # acceptable url chars in parens
       [-A-Z0-9+&@#/%=~_|$?!:,.])*            # acceptable url chars
       (?:\([-A-Z0-9+&@#/%=~_|$?!:,.]*\)|     # acceptable url chars in parens
       [A-Z0-9+&@#/%=~_|$])                   # acceptable url chars  
    """, re.VERBOSE)



import os
import random
from collections import Counter

import pytest
from scrapy.http import HtmlResponse, Request

from advertools.spider import (
    _IMG_ATTRS,
    _extract_images,
    _json_to_dict,
    _numbered_duplicates,
    _split_long_urllist,
    crawl,
)

jsonobj = {
    "@context": "http://schema.org",
    "@type": "NewsMediaOrganization",
    "name": "The New York Times",
    "logo": {
        "@context": "http://schema.org",
        "@type": "ImageObject",
        "url": "https://static01.nyt.com/images/misc/NYT_logo_rss_250x40.png",
        "height": 40,
        "width": 250,
    },
    "url": "https://www.nytimes.com/",
    "@id": "https://www.nytimes.com/#publisher",
    "diversityPolicy": "https://www.nytco.com/diversity-and-inclusion-at-the-new-york-times/",
    "ethicsPolicy": "https://www.nytco.com/who-we-are/culture/standards-and-ethics/",
    "masthead": "https://www.nytimes.com/interactive/2019/admin/the-new-york-times-masthead.html",
    "foundingDate": "1851-09-18",
    "sameAs": "https://en.wikipedia.org/wiki/The_New_York_Times",
}


def response_from_file(filename, url="https://example.com"):
    request = Request(url=url)
    filepath = os.path.abspath(filename)
    with open(filepath) as file:
        file_content = file.read()
        response = HtmlResponse(
            url=url, request=request, body=file_content, encoding="utf-8"
        )
    return response


def test_numbered_duplicates_returns_same_num_of_items():
    l1 = ["jan", "feb", "jan", "jan", "feb", "mar", "apr"]
    l2 = ["one", "one", "one"]
    l3 = ["one", "two", "three", "three", "three", "two"]
    for lst in [l1, l2, l3]:
        result = _numbered_duplicates(lst)
        assert len(result) == len(lst)


def test_numbered_duplicates_returns_correct_items():
    numbers = ["one", "two", "three", "four"]
    for i in range(1000):
        sample = random.choices(numbers, k=7)
        result = _numbered_duplicates(sample)
        result_split = [num.rsplit("_", 1)[0] for num in result]
        assert Counter(sample) == Counter(result_split)


def test_json_to_dict_returns_dict():
    result = _json_to_dict(jsonobj)
    assert isinstance(result, dict)
    assert all("jsonld_" in key for key in result)


def test_json_to_dict_contains_number():
    for i in range(1, 6):
        result = _json_to_dict(jsonobj, i)
        to_test = "jsonld_" + str(i) + "_"
        assert isinstance(result, dict)
        assert all(to_test in key for key in result)


def test_crawl_raises_on_wrong_file_extension():
    with pytest.raises(ValueError):
        crawl("https://example.com", "myfile.wrong", allowed_domains="example.com")


def test_split_long_urllist_correct_lengths():
    testlist = ["one", "two", "three", "four", "five", "six", "seven", "eight"]
    result = _split_long_urllist(testlist, max_len=15)
    assert sum(len(x) for x in result) == len(testlist)
    assert all(len(x) < 15 for x in result)


def test_extract_images(html_file="tests/data/crawl_testing/test_images.html"):
    response = response_from_file(html_file)
    imgs_dict = _extract_images(response)
    for k, v in imgs_dict.items():
        assert len(v.split("@@")) == 3
    assert imgs_dict["img_src"] == "@@https://example.com/image.png@@"



import os

import pytest

from advertools.twitter import *
from advertools.twitter import FUNCTIONS
from advertools.twitter import _get_counts

app_key = os.environ.get('APP_KEY')
app_secret = os.environ.get('APP_SECRET')
oauth_token = os.environ.get('OAUTH_TOKEN')
oauth_token_secret = os.environ.get('OAUTH_TOKEN_SECRET')

auth_params = {
    'app_key': app_key,
    'app_secret': app_secret,
    'oauth_token': oauth_token,
    'oauth_token_secret': oauth_token_secret,
}


set_auth_params(**auth_params)

skip_api_tests = pytest.mark.skipif(os.environ.get('ADV_TEST_OFFLINE'),
                                    reason='Run all except API dependents')

def test_get_counts():
    for i in [None, 13, 70, 100, 101, 200, 578]:
        result = _get_counts(i, 100)
        assert sum(result) == i if i is not None else 100
        assert 0 not in result


def test_set_auth_params():
    set_auth_params(**auth_params)
    for func in FUNCTIONS:
        assert func.get_auth_params()['app_key'] == app_key
        assert func.get_auth_params()['app_secret'] == app_secret
        assert func.get_auth_params()['oauth_token'] == oauth_token
        assert func.get_auth_params()['oauth_token_secret'] == oauth_token_secret


@skip_api_tests
def test_get_application_rate_limit_status():
    for truth in [True, False]:
        result = get_application_rate_limit_status(truth)
        assert type(result) == pd.core.frame.DataFrame
        assert 'endpoint' in result


@skip_api_tests
def test_get_available_trends():
    result = get_available_trends()
    assert type(result) == pd.core.frame.DataFrame
    assert 'woeid' in result


@skip_api_tests
def test_get_favorites():
    result = get_favorites(screen_name='twitter', count=5,
                           tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'tweet_full_text' in result
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])


@skip_api_tests
def test_get_followers_ids():
    result = get_followers_ids(screen_name='nytimes', count=5)
    assert len(result['ids']) <= 5


@skip_api_tests
def test_get_followers_list():
    result = get_followers_list(screen_name='eliasdabbas', count=5,
                                skip_status=None)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5 
    assert 'screen_name' in result


@skip_api_tests
def test_get_friends_ids():
    result = get_friends_ids(screen_name='twitter', count=5)
    assert len(result['ids']) <= 5 


@skip_api_tests
def test_get_friends_list():
    result = get_friends_list(screen_name='shakira', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'screen_name' in result


@skip_api_tests
def test_get_home_timeline():
    result = get_home_timeline(count=5, tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'tweet_full_text' in result


@skip_api_tests
def test_get_list_members():
    result = get_list_members(slug='nyt-journalists',
                              owner_screen_name='nytimes', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'screen_name' in result


@skip_api_tests
def test_get_list_memberships():
    result = get_list_memberships(screen_name='ThePSF', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'list_full_name' in result


@skip_api_tests
def test_get_list_statuses():
    result = get_list_statuses(slug='nyt-journalists',
                               owner_screen_name='nytimes', count=5,
                               tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'tweet_full_text' in result
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])


@skip_api_tests
def test_get_list_subscribers():
    result = get_list_subscribers(slug='nyt-journalists',
                                  owner_screen_name='nytimes', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5


@skip_api_tests
def test_get_list_subscriptions():
    result = get_list_subscriptions(screen_name='twitter', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5


@skip_api_tests
def test_get_mentions_timeline():
    result = get_mentions_timeline(count=5, tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'tweet_full_text' in result
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])


@skip_api_tests
def test_get_place_trends():
    result = get_place_trends(ids=[23424950, 766273])
    assert type(result) == pd.core.frame.DataFrame
    assert result['location'].str.contains('Madrid|Spain').all()
    result_single_id = get_place_trends(ids=1)
    assert type(result_single_id) == pd.core.frame.DataFrame


@skip_api_tests
def test_get_retweeters_ids():
    result = get_retweeters_ids(id=1032676609902428160, count=5)
    assert len(result['ids']) <= 5


@skip_api_tests
def test_get_retweets():
    result = get_retweets(id=1032676609902428160, tweet_mode='extended', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])


@skip_api_tests
def test_get_supported_languages():
    result = get_supported_languages()
    assert type(result) == pd.core.frame.DataFrame
    assert 'code' in result


@skip_api_tests
def test_get_user_timeline():
    result = get_user_timeline(screen_name='bbc', count=5,
                               tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert 'tweet_full_text' in result
    assert 'user_id' in result
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])
    long_result = get_user_timeline(screen_name='eliasdabbas', count=2000,
                                    tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame


@skip_api_tests
def test_lookup_status():
    result = lookup_status(id=[1032676609902428160,
                               1031574983095533568],
                           tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])


@skip_api_tests
def test_lookup_user():
    result = lookup_user(screen_name=['twitter', 'nytimes', 'shakira'])
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 3


@skip_api_tests
def test_retweeted_of_me():
    result = retweeted_of_me(count=5, tweet_mode='extended')
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])


@skip_api_tests
def test_search():
    result = search(q='basketball', lang='en', count=5, tweet_mode='extended')
    result_compat = search(q='basketball', lang='en', count=5,
                           tweet_mode='compat')
    assert type(result) == pd.core.frame.DataFrame
    assert type(result_compat) == pd.core.frame.DataFrame
    assert len(result) <= 5
    assert len(result_compat) <= 5
    assert 'tweet_full_text' in result
    assert 'tweet_text' in result_compat
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])
    assert all([embd_uid == uid for embd_uid, uid in
                zip([x['id'] for x in result['tweet_user']],
                    result['user_id'])])


@skip_api_tests
def test_search_no_error_on_empty_result():
    result = search(q='thisqueryhasnoresult!@#$%^', lang='ar',
                    count=200, tweet_mode='extended')
    assert isinstance(result, pd.core.frame.DataFrame)
    assert len(result) == 0


@skip_api_tests
def test_search_users():
    result = search_users(q='finance', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5


@skip_api_tests
def test_show_lists():
    result = show_lists(screen_name='bbc')
    assert type(result) == pd.core.frame.DataFrame


@skip_api_tests
def test_show_owned_lists():
    result = show_owned_lists(screen_name='nytimes', count=5)
    assert type(result) == pd.core.frame.DataFrame
    assert len(result) <= 5
